from typing import *

import os
import importlib
import inspect


class QxInspectUtils(object):
  @staticmethod
  def get_method_class(method: Callable) -> Optional[Type]:
    ret = None

    if method.__class__.__name__ in ('function', 'method') and method.__qualname__ != method.__name__:
      try:
        mod = importlib.import_module(method.__module__)
        cls_name = method.__qualname__[:-len(method.__name__) - 1]
        cls = getattr(mod, cls_name, None)
        if cls is not None and isinstance(cls, type):
          ret = cls
      except ModuleNotFoundError as e:
        import traceback
        traceback.print_exc()
        pass
    return ret

  @staticmethod
  def is_static_method(method: Callable) -> bool:
    args = inspect.getfullargspec(method).args
    return not len(args) or args[0] not in ('cls', 'self')

  @staticmethod
  def is_class_method(method: Callable) -> bool:
    args = inspect.getfullargspec(method).args
    return len(args) and args[0] == 'cls'

  @staticmethod
  def is_inst_method(method: Callable) -> bool:
    args = inspect.getfullargspec(method).args
    return len(args) and args[0] == 'self'

  @staticmethod
  def fi() -> str:
    return inspect.stack()[-1].filename

  @staticmethod
  def fib() -> str:
    return os.path.basename(inspect.stack()[-1].filename)



import functools
import importlib
import os
from typing import *

from cached_property import cached_property

from quicly.model import QxAttr, QxOptional
from quicly.decorator import QxDecoratorHook, QxDecorator
from quicly.conf import QxConf, QxGlobalConf
from quicly.urlutils import QxUrl
from quicly.result import QxResult, QxError, QxFinish, QxHttpResult, QxTemplate
from quicly import jsonutils as json
from quicly.patterns import PATTERN_URL_QUERY
from quicly.inspectutils import QxInspectUtils
from quicly.session import QxSession
from quicly.logutils import QxLogging

import re
import copy
import threading
import uuid as _uuid
import inspect
import bson
import pickle
import pyDes
from collections import OrderedDict
from datetime import datetime
import uuid
import yaml

logging = QxLogging.get_logger('QxServer')

_RPC_METHODS = {'RPC'}
_HTTP_METHODS = {'GET', 'PUT', 'PATCH', 'DELETE', 'POST', 'HEAD', 'OPTIONS'}
_METHODS = _RPC_METHODS.union(_HTTP_METHODS)
_VAR_DEF_PATTERN = re.compile(r'{[_A-Z]+[_0-9A-Z]*(:[^}]*)?}', re.I)
_VAR_VAL_PATTERN = '[^/]*'


class QxCaseInsensitiveDict(OrderedDict):
  _mapping = dict()

  def __init__(self, *al, **kw):
    super(QxCaseInsensitiveDict, self).__init__(*al, **kw)
    for k, v in super(QxCaseInsensitiveDict, self).items():
      self._mapping[k.lower()] = k

  def get(self, k: str, default: Any = None) -> Any:
    kk = self._mapping.get(k.lower())
    if kk:
      ret = super(QxCaseInsensitiveDict, self).get(kk, default)
    else:
      ret = default
    return ret

  def __getitem__(self, k: str):
    return self.get(k)

  def __setitem__(self, k: str, v: Any):
    self._mapping[k.lower()] = k
    super(QxCaseInsensitiveDict, self).__setitem__(k, v)


QxHeaders = QxCaseInsensitiveDict
QxCookies = QxCaseInsensitiveDict


class QxRequest(object):
  def __init__(self, method: str, url: str, headers: Optional[Dict[str, str]] = None, cookies: Optional[Dict[str, str]] = None, body: Optional[bytes] = None, form: Optional[Dict[str, str]] = None, files: List = None, request_uuid: str = None, request_time: Optional[datetime] = None):
    self._method = method.upper()
    self._url = QxUrl(url)
    self._headers = QxHeaders(headers.items() if isinstance(headers, dict) else [])
    self._cookies = QxCookies(cookies.items() if isinstance(cookies, dict) else [])
    self._body = body if isinstance(body, (bytes, str)) else b''
    self._data = [self._body]
    self._args = dict()
    self._form = form if isinstance(form, dict) else {}
    self._files = files if isinstance(files, list) else []
    self._request_uuid = request_uuid if request_uuid else str(uuid.uuid4())
    self._request_time = request_time if request_time else datetime.utcnow()
    self._context = None
    self._session = None
    self._user = None

  @property
  def method(self) -> str:
    return self._method

  @property
  def url(self) -> QxUrl:
    return self._url

  @property
  def headers(self) -> Dict[str, str]:
    return self._headers

  @property
  def cookies(self) -> Dict[str, str]:
    return self._cookies

  @property
  def body(self) -> Optional[bytes]:
    return self._body

  @property
  def data(self) -> Any:
    return self._data[-1] if self._data else None

  @data.setter
  def data(self, data: Any):
    self._data.append(data)

  @property
  def args(self) -> Dict:
    return self._args

  @args.setter
  def args(self, args):
    self._args = args

  @property
  def form(self) -> Dict[str, str]:
    return self._form

  @property
  def files(self) -> List:
    return self._files

  @property
  def request_uuid(self):
    return self._request_uuid

  @property
  def request_time(self):
    return self._request_time

  @property
  def context(self) -> Any:
    return self._context

  @context.setter
  def context(self, context: Any):
    self._context = context

  @property
  def session(self) -> QxSession:
    return self._session

  @session.setter
  def session(self, session: QxSession):
    self._session = session

  @property
  def user(self) -> Optional[Dict]:
    return self._user

  @user.setter
  def user(self, user: Optional[Dict]):
    self._user = user

  def get_content_type(self):
    items = [x.strip().lower() for x in self.headers.get('Content-Type', '').split(';')]
    content_type = items[0]
    if len(items) > 1 and items[1].startswith('charset:'):
      charset = items[1].split(':', maxsplit=1)[-1]
    else:
      charset = None
    return content_type, charset

  def get_accept(self):
    accept = set([x.strip().lower() for x in self.headers.get('Accept', '').split(';')])
    charset = self.headers.get('Accept-Charset', None)
    return accept, charset

  def get_content_encryption(self):
    items = [x.strip() for x in self.headers.get('X-Qx-Content-Encryption', '').split(';')]
    if len(items) < 2:
      items.append('')
    return tuple(items)

  def get_accept_encryption(self):
    return self.headers.get('X-Qx-Accept-Encryption', '')

  def parse_args(self):
    args = dict()

    for arg_t in (self.cookies.get('qx-args'), self.headers.get('x-qx-args')):
      if not isinstance(arg_t, str):
        continue

      try:
        arg_t_d = json.loads(arg_t)
        if isinstance(arg_t_d, dict):
          for k, v in arg_t_d.items():
            args[k] = v
      except (pickle.PickleError, json.JSONDecodeError, TypeError, ValueError):
        if PATTERN_URL_QUERY.fullmatch(arg_t) and arg_t.count('"') == 0 and arg_t.count("'") == 0 and (arg_t.count('&') > 0 or arg_t.count('=') > 0):
          for k, v in QxUrl.parse_query_str(arg_t).items():
            args[k] = v

    if isinstance(self.url.query, dict):
      for k, v in self.url.query.items():
        args[k] = v

    if isinstance(self.form, dict):
      for k, v in self.form.items():
        args[k] = v

    if isinstance(self.data, dict):
      for k, v in self.data.items():
        args[k] = v

    self.args = args


class QxResponse(object):
  def __init__(self, request: QxRequest, code: int = 0, headers: Optional[Dict[str, str]] = None, cookies: Optional[Dict[str, str]] = None, body: Optional[bytes] = ''):
    self._request = request
    self._code = code
    self._headers = QxHeaders(headers.items() if isinstance(headers, dict) else [])
    self._cookies = QxCookies(cookies.items() if isinstance(cookies, dict) else [])
    self._body = body
    self._data = [body]

  @property
  def request(self) -> QxRequest:
    return self._request

  @property
  def code(self) -> int:
    return self._code

  @property
  def headers(self) -> Dict[str, str]:
    return self._headers

  @property
  def cookies(self) -> Dict[str, str]:
    return self._cookies

  @property
  def body(self) -> Optional[bytes]:
    return self._body

  @property
  def data(self) -> Any:
    return self._data[-1] if self._data else None

  @data.setter
  def data(self, data: Any):
    self._data.append(data)


class QxHandlerFunc(object):
  def __init__(self, func: Callable, args: Optional[Dict[str, QxAttr]], result: Optional[QxAttr] = None, auth: bool = False):
    self._func = func if callable(func) else lambda: QxError(QxError.CODE_501_NOT_IMPLEMENTED)
    self._args = args if isinstance(args, dict) else dict()
    self._result = result
    self._auth = auth
    self._func_cached = None

  @property
  def func(self) -> Callable:
    return self._func

  @property
  def args(self) -> Optional[Dict[str, QxAttr]]:
    return self._args

  @property
  def result(self) -> Optional[QxAttr]:
    return self._result

  @property
  def auth(self) -> bool:
    return self._auth

  def _check_args(self, args: Dict[str, Any]) -> Optional[QxError]:
    err = None

    if isinstance(args, dict):
      for k, v in self.args.items():
        if k in args or isinstance(v, QxOptional):
          if not v.check(args.get(k)):
            err = QxError(code=QxError.CODE_400_BAD_REQUEST, reason='INVALID_ARGS', data={
              'name': k,
            })
            break
        else:
          err = QxError(code=QxError.CODE_400_BAD_REQUEST, reason='MISSING_REQUIRED_ARGS', data={
            'name': k,
          })
    else:
      err = QxError(code=QxError.CODE_400_BAD_REQUEST)

    return err

  def _process_args(self, args: Dict[str, Any]) -> Dict[str, Any]:
    ret = dict()

    if isinstance(args, dict):
      for k, v in args.items():
        if isinstance(self.args, dict) and k in self.args and isinstance(self.args[k], QxAttr):
          v = self.args[k].value(v)
        ret[k] = copy.copy(v)

    return ret

  def _check_result(self, result: Any) -> bool:
    ret = True

    if isinstance(self.result, QxAttr) and not self.result.check(result):
      ret = False

    return ret

  def _process_result(self, result: Any) -> Any:
    ret = None

    if isinstance(self.result, QxAttr):
      if self.result.check(result):
        result = self.result.value(result)
        ret = copy.copy(result)
    else:
      ret = result

    return ret

  def _get_func(self) -> Callable:
    if callable(self._func_cached):
      func = self._func_cached
    else:
      func = self._func

      if QxInspectUtils.is_class_method(func):
        cls = QxInspectUtils.get_method_class(func)
        func = functools.partial(func, cls)
      elif QxInspectUtils.is_inst_method(func):
        cls = QxInspectUtils.get_method_class(func)
        inst_k = '__qx_handler_inst__'
        inst = getattr(cls, inst_k, None)
        if inst is None or not isinstance(inst, cls):
          inst = cls()
          setattr(cls, inst_k, inst)
        func = functools.partial(func, inst)
      self._func_cached = func
    return func

  def __call__(self, request: QxRequest) -> QxResult:
    if callable(self._func):
      if self.auth and not request.user:
        err = QxError(code=QxError.CODE_401_UNAUTHORIZED)
      else:
        err = self._check_args(request.args)

      if not isinstance(err, Exception):
        request.args = self._process_args(request.args)

        try:
          func = self._get_func()

          kw_full = dict(
            request=request,
            args=request.args,
          )

          func_args = inspect.getfullargspec(func)

          kw = kw_full if func_args.varkw else dict([x for x in kw_full.items() if x[0] in func_args.args])

          result = func(**kw)
        except QxFinish as f:
          result = f
        except QxError as e:
          result = e
        except QxResult as r:
          result = r
        except Exception as e:
          import traceback
          traceback.print_exc()
          result = QxError(data=e)

        if not isinstance(result, QxResult):
          if isinstance(result, Exception):
            import traceback
            traceback.print_exc()
            result = QxError(data=result)
          else:
            result = QxFinish(data=result)
      else:
        result = err if isinstance(err, QxError) else QxError(code=QxError.CODE_400_BAD_REQUEST)
    else:
      result = QxError(code=QxError.CODE_501_NOT_IMPLEMENTED)

    try:
      assert isinstance(result, QxResult)
    except AssertionError as e:
      result = QxError(data=e)

    return result


class QxHandlerMeta(object):
  def __init__(self, path: str, methods: Union[str, List[str], Tuple[str], Set[str]] = None):
    self._path, self._pattern, self._varlist = self._process_path(path)
    self._methods = self._process_methods(methods)

  @property
  def path(self) -> str:
    return self._path

  @property
  def pattern(self) -> Optional[re.Pattern]:
    return self._pattern

  @property
  def varlist(self) -> Tuple[str]:
    return self._varlist

  @property
  def methods(self) -> Set[str]:
    return self._methods

  @staticmethod
  def _process_path(path: str) -> Tuple[str, Optional[re.Pattern], Tuple[str]]:
    matches = list(_VAR_DEF_PATTERN.finditer(path))

    pattern = None
    varlist = []

    if len(matches):
      p = ''

      i = 0
      for m in matches:
        j, k = m.span()

        var_fields = path[j + 1:k - 1].split(':', maxsplit=1)
        var_name = var_fields[0].strip()
        var_pattern = var_fields[1].strip() if len(var_fields) >= 2 else _VAR_VAL_PATTERN

        p += path[i:j]
        p += f'({var_pattern})'

        varlist.append(var_name)

        i = k

      p += path[i:]

      pattern = re.compile(p)

    varlist = tuple(varlist)

    return path, pattern, varlist

  @staticmethod
  def _process_methods(methods: Union[str, List[str], Tuple[str], Set[str]]) -> Set[str]:
    if isinstance(methods, str):
      methods = methods.upper()
      if methods in _METHODS:
        ret = {methods}
      elif methods == 'RPC':
        ret = _RPC_METHODS
      elif methods == 'HTTP':
        ret = _HTTP_METHODS
      else:
        ret = _METHODS
    elif isinstance(methods, (list, tuple, set)):
      ret = set()
      for m in methods:
        if not isinstance(m, str):
          continue
        m = m.upper()

        if m in _METHODS:
          ret.add(m)
        elif m == 'RPC':
          ret += _RPC_METHODS
        elif m == 'HTTP':
          ret += _HTTP_METHODS
        elif m == '*':
          ret += _METHODS

      if not ret:
        ret = _METHODS
    else:
      ret = _METHODS

    return ret


class QxHandler(object):
  def __init__(self, func: QxHandlerFunc, meta: QxHandlerMeta):
    self._func = func
    self._meta = meta

  @property
  def func(self) -> QxHandlerFunc:
    return self._func

  @property
  def meta(self) -> QxHandlerMeta:
    return self._meta


class QxHandlers(object):
  def __init__(self):
    self._handlers = OrderedDict()  # type: Dict[str, List[QxHandler]]

  @property
  def handlers(self) -> Dict[str, List[QxHandler]]:
    return self._handlers

  def clear(self):
    self._handlers.clear()

  def register(self, handler: QxHandler):
    self._handlers.setdefault(handler.meta.path, [])
    self._handlers[handler.meta.path].append(handler)

  def match(self, path: str, method: Optional[str] = None) -> List[Tuple[QxHandler, List[Tuple[str, str]]]]:
    ret = list()

    for handler in self.handlers.get(path, []):  # type: QxHandler
      if method and method not in handler.meta.methods:
        continue
      ret.append((handler, []))

    if not len(ret):
      for handlers in self.handlers.values():  # type: List[QxHandler]
        for handler in handlers:  # type: QxHandler
          if method and method not in handler.meta.methods:
            continue

          if isinstance(handler, QxHandler) and isinstance(handler.meta.pattern, re.Pattern) and handler.meta.pattern.fullmatch(path):
            args = list()

            args_t = handler.meta.pattern.findall(path)[0]
            for i in range(len(args_t)):
              args.append((
                handler.meta.varlist[i],
                args_t[i],
              ))

            ret.append((handler, args))

    return ret


class QxDecoder(object):
  def can_decode(self, request: QxRequest) -> bool:
    raise NotImplementedError()

  def decode(self, request: QxRequest) -> QxRequest:
    raise NotImplementedError()


class QxDecoderLayer(QxDecoder):
  def __init__(self, *decoders):
    self._decoders = [x for x in decoders if isinstance(x, QxDecoder)]

  @property
  def decoders(self) -> List[QxDecoder]:
    return self._decoders

  def can_decode(self, request: QxRequest) -> bool:
    return True

  def decode(self, request: QxRequest) -> QxRequest:
    for decoder in self.decoders:
      if decoder.can_decode(request):
        request = decoder.decode(request)
        break
    return request


class QxDecoderLayers(QxDecoder):
  def __init__(self, *layers):
    self._layers = [x for x in layers if isinstance(x, QxDecoderLayer)]

  @property
  def layers(self) -> List[QxDecoderLayer]:
    return self._layers

  def can_decode(self, request: QxRequest) -> bool:
    return True

  def decode(self, request: QxRequest) -> QxRequest:
    for layer in self.layers:
      if layer.can_decode(request):
        request = layer.decode(request)
    return request


class QxEncoder(object):
  def can_encode(self, response: QxResponse) -> bool:
    raise NotImplementedError()

  def encode(self, response: QxResponse) -> QxResponse:
    raise NotImplementedError()


class QxEncoderLayer(QxEncoder):
  def __init__(self, *encoders):
    self._encoder = [x for x in encoders if isinstance(x, QxEncoder)]

  @property
  def encoders(self) -> List[QxEncoder]:
    return self._encoder

  def can_encode(self, response: QxResponse) -> bool:
    return True

  def encode(self, response: QxResponse) -> QxResponse:
    for encoder in self.encoders:
      if encoder.can_encode(response):
        response = encoder.encode(response)
        break
    return response


class QxEncoderLayers(QxEncoder):
  def __init__(self, *layers):
    self._layers = [x for x in layers if isinstance(x, QxEncoderLayer)]

  @property
  def layers(self) -> List[QxEncoderLayer]:
    return self._layers

  def can_encode(self, response: QxResponse) -> bool:
    return True

  def encode(self, response: QxResponse) -> QxResponse:
    for layer in self.layers:
      if layer.can_encode(response):
        response = layer.encode(response)
        break
    return response


class QxCoder(QxEncoder, QxDecoder):
  _lock = threading.Lock()
  _inst = None

  def __new__(cls, *al, **kw):
    if not isinstance(cls._inst, cls):
      with cls._lock:
        if not isinstance(cls._inst, cls):
          cls._inst = super().__new__(cls, *al, **kw)
    return cls._inst

  def can_encode(self, response: QxResponse) -> bool:
    raise NotImplementedError()

  def encode(self, response: QxResponse) -> QxResponse:
    raise NotImplementedError()

  def can_decode(self, request: QxRequest) -> bool:
    raise NotImplementedError()

  def decode(self, request: QxRequest) -> QxRequest:
    raise NotImplementedError()


class QxDESCoder(QxCoder):
  @staticmethod
  def _mk_des(p: str):
    return pyDes.des(key=p, mode=pyDes.ECB, IV=None, pad=None, padmode=pyDes.PAD_NORMAL)

  def can_encode(self, response: QxResponse) -> bool:
    return response.request.get_accept_encryption().lower() == 'des'

  def encode(self, response: QxResponse) -> QxResponse:
    p = str(_uuid.uuid4())
    des = self._mk_des(p)
    data = des.encrypt(response.data)
    response.data = data
    return response

  def can_decode(self, request: QxRequest) -> bool:
    e, p = request.get_content_encryption()
    return e.lower() == 'des' and p

  def decode(self, request: QxRequest) -> QxRequest:
    _, p = request.get_content_encryption()
    des = self._mk_des(p)
    data = des.decrypt(request.data)
    request.data = data
    return request


class QxPickleCoder(QxCoder):
  def can_encode(self, response: QxResponse) -> bool:
    accept, _ = response.request.get_accept()
    return 'application/pickle' in accept

  def encode(self, response: QxResponse) -> QxResponse:
    data = pickle.dumps(response.data)
    response.data = data
    response.headers['Content-Type'] = 'application/octet-stream'
    response.headers['Content-Length'] = str(len(data))
    return response

  def can_decode(self, request: QxRequest) -> bool:
    if isinstance(request.data, bytes) and len(request.data) >= 2 and request.data[0] == 128 and request.data[1] == 4:
      try:
        pickle.loads(request.data)
        ret = True
      except (pickle.PickleError, TypeError, ValueError):
        ret = False
    else:
      ret = False
    return ret

  def decode(self, request: QxRequest) -> QxRequest:
    data = pickle.loads(request.data)
    request.data = data
    return request


class QxBsonCoder(QxCoder):
  def can_encode(self, response: QxResponse) -> bool:
    accept, _ = response.request.get_accept()
    return 'application/bson' in accept and isinstance(response.data, dict)

  def encode(self, response: QxResponse) -> QxResponse:
    data = bson.encode(response.data)
    response.data = data
    response.headers['Content-Type'] = 'application/octet-stream'
    response.headers['Content-Length'] = str(len(data))
    return response

  def can_decode(self, request: QxRequest) -> bool:
    if isinstance(request.data, bytes) and len(request.data):
      try:
        bson.decode(request.data)
        ret = True
      except (bson.errors.BSONError, TypeError, ValueError):
        ret = False
    else:
      ret = False
    return ret

  def decode(self, request: QxRequest) -> QxRequest:
    data = bson.decode(request.data)
    request.data = data
    return request


class QxYamlCoder(QxCoder):
  def can_encode(self, response: QxResponse) -> bool:
    accept, _ = response.request.get_accept()

    ret = False

    if 'application/yaml' in accept:
      ret = True

    return ret

  def encode(self, response: QxResponse) -> QxResponse:
    _, charset = response.request.get_accept()
    if not charset:
      charset = QxGlobalConf().charset
    data = yaml.safe_dump(response.data)
    if isinstance(data, str):
      data = data.encode(charset)
    response.data = data
    response.headers['Content-Type'] = 'application/yaml'
    response.headers['Content-Length'] = str(len(data))
    return response

  def can_decode(self, request: QxRequest) -> bool:
    if isinstance(request.data, (str, bytes)) and len(request.data):
      try:
        yaml.safe_load(request.data)
        ret = True
      except (yaml.YAMLError, TypeError, ValueError):
        ret = False
    else:
      ret = False
    return ret

  def decode(self, request: QxRequest) -> QxRequest:
    _, charset = request.get_content_type()
    if not charset:
      charset = QxGlobalConf().charset
    data = request.data
    if isinstance(data, bytes):
      data = data.decode(charset)
    request.data = yaml.safe_load(data)
    return request


class QxJsonCoder(QxCoder):
  def can_encode(self, response: QxResponse) -> bool:
    accept, _ = response.request.get_accept()

    ret = False

    if 'application/json' in accept:
      ret = True
    elif isinstance(response.data, dict) and 'code' in response.data and 'reason' in response.data and 'message' in response.data and 'data' in response.data:
      ret = True

    return ret

  def encode(self, response: QxResponse) -> QxResponse:
    _, charset = response.request.get_accept()
    if not charset:
      charset = QxGlobalConf().charset
    data = json.dumps(response.data, indent=0, ensure_ascii=False)
    if isinstance(data, str):
      data = data.encode(charset)
    response.data = data
    response.headers['Content-Type'] = 'application/json'
    response.headers['Content-Length'] = str(len(data))
    return response

  def can_decode(self, request: QxRequest) -> bool:
    if isinstance(request.data, (str, bytes)) and len(request.data):
      try:
        json.loads(request.data)
        ret = True
      except (json.JSONDecodeError, TypeError, ValueError):
        ret = False
    else:
      ret = False
    return ret

  def decode(self, request: QxRequest) -> QxRequest:
    _, charset = request.get_content_type()
    if not charset:
      charset = QxGlobalConf().charset
    data = request.data
    if isinstance(data, bytes):
      data = data.decode(charset)
    request.data = json.loads(data)
    return request


class QxServerHook(object):
  def render_template(self, name, **kw) -> Union[str, bytes]:
    return ''

  def init_server(self) -> NoReturn:
    pass

  def load_context(self, request: QxRequest) -> Optional[Dict]:
    pass

  def load_session(self, request: QxRequest) -> Optional[Dict]:
    pass

  def load_user(self, request: QxRequest) -> Optional[Dict]:
    pass

  def before_request(self, request: QxRequest) -> Any:
    pass

  def on_result(self, request: QxRequest, result: QxResult) -> Optional[QxResult]:
    pass

  def after_request(self, request: QxRequest, response: QxResponse) -> QxResponse:
    pass


class QxServer(object):
  _global_handlers = OrderedDict()

  @classmethod
  def get_handlers(cls, name: str):
    handlers = cls._global_handlers.get(name)
    if not isinstance(handlers, QxHandlers):
      handlers = QxHandlers()
      cls._global_handlers[name] = handlers
    return handlers

  ################################################################################

  _decoders = QxDecoderLayers(
    QxDecoderLayer(  # 解密
      QxDESCoder(),
    ),
    QxDecoderLayer(  # 解码
      QxPickleCoder(),
      QxBsonCoder(),
      QxYamlCoder(),
      QxJsonCoder(),
    ),
  )

  _encoders = QxEncoderLayers(
    QxEncoderLayer(  # 编码
      QxPickleCoder(),
      QxBsonCoder(),
      QxYamlCoder(),
      QxJsonCoder(),
    ),
    QxEncoderLayer(  # 加密
      QxDESCoder(),
    ),
  )

  ################################################################################

  def __init__(self, name: str = None, hook: QxServerHook = None):
    self._name = name
    self._hook = hook
    self._conf = QxConf(parent=QxGlobalConf())

  @property
  def name(self) -> str:
    return self._name

  @property
  def hook(self) -> QxServerHook:
    return self._hook

  @property
  def conf(self) -> QxConf:
    return self._conf

  @cached_property
  def handlers(self) -> QxHandlers:
    return self.get_handlers(self.name)

  @cached_property
  def decoders(self) -> QxDecoderLayers:
    return self._decoders

  @cached_property
  def encoders(self) -> QxEncoderLayers:
    return self._encoders

  @staticmethod
  def get_http_header_server_name():
    return 'Quicly'

  def _load_handlers(self, home: str, name: str):
    if not os.path.isdir(home):
      return

    try:
      importlib.import_module(name)
      logging.info(f'{name} : loaded')
    except ImportError:
      logging.error(f'{name} : failed')

    for item in os.listdir(home):
      item = os.path.basename(item)
      item_t = os.path.splitext(item)[0]
      if item.endswith('Handlers') and os.path.isdir(os.path.join(home, item)):
        self._load_handlers(os.path.join(home, item), f'{name}.{item}')
      elif item_t.endswith('Handler') and os.path.isfile(os.path.join(home, item)):
        handler_name = f'{name}.{item_t}'
        try:
          importlib.import_module(f'{handler_name}')
          logging.info(f'{handler_name} : loaded')
        except ImportError:
          logging.error(f'{handler_name} : failed')

  def init(self):
    self._load_handlers(os.path.join(os.getcwd(), 'handlers'), 'handlers')

    if isinstance(self.hook, QxServerHook):
      self.hook.init_server()

  def _dispatch(self, request: QxRequest) -> Any:
    ret = QxError(code=QxError.CODE_404_NOT_FOUND)

    for handler, args in self.handlers.match(request.url.path):
      ret = QxError(code=QxError.CODE_405_METHOD_NOT_ALLOWED)

      if request.method in handler.meta.methods:
        for k, v in args:
          request.args[k] = v

        if isinstance(self.hook, QxServerHook):
          request.user = self.hook.load_user(request)

        ret = handler.func(request)
        break

    return ret

  def handle(self, request: QxRequest):
    request = self.decoders.decode(request)

    request.parse_args()

    try:
      if isinstance(self.hook, QxServerHook):
        request.context = self.hook.load_context(request)
        request.session = self.hook.load_session(request)

        if isinstance(request.session, QxSession):
          with request.session:
            result = self.hook.before_request(request)
            if result is None:
              result = self._dispatch(request)
        else:
          result = self.hook.before_request(request)
          if result is None:
            result = self._dispatch(request)
      else:
        result = self._dispatch(request)
    except QxFinish as f:
      result = f
    except QxError as e:
      result = e
    except QxResult as r:
      result = r
    except Exception as e:
      import traceback
      traceback.print_exc()
      result = QxError(data=e)

    if not isinstance(result, QxResult):
      if isinstance(result, Exception):
        result = QxError(data=result)
      else:
        result = QxFinish(data=result)

    try:
      assert isinstance(result, QxResult)
    except AssertionError as e:
      result = QxError(data=e)

    if isinstance(self.hook, QxServerHook):
      result_t = self.hook.on_result(request, result)
      if isinstance(result_t, QxResult):
        result = result_t

    if isinstance(result, QxHttpResult):
      if isinstance(result, QxTemplate):
        if isinstance(self.hook, QxServerHook):
          response = QxResponse(
            request=request,
            code=result.code,
            body=self.hook.render_template(**result.content),
            headers=result.headers,
            cookies=result.cookies,
          )
          response.headers.setdefault('Content-Type', 'text/html')
        else:
          response = QxResponse(
            request=request,
            body=QxError(code=QxError.CODE_501_NOT_IMPLEMENTED).to_dict(),
          )
      else:
        response = QxResponse(
          request=request,
          code=result.code,
          body=result.content,
          headers=result.headers,
          cookies=result.cookies,
        )
        response.headers.setdefault('Content-Type', 'text/html')
    else:
      response = QxResponse(
        request=request,
        body=result.to_dict(),
      )

    response = self.encoders.encode(response)

    response.headers['Server'] = self.get_http_header_server_name()

    if isinstance(self.hook, QxServerHook):
      response_t = self.hook.after_request(request, response)
      response = response_t if isinstance(response_t, QxResponse) else response

    return response

################################################################################


class QxServerDecoratorHook(QxDecoratorHook):
  def _check(self, target: Any) -> bool:
    return callable(target) or isinstance(target, (classmethod, staticmethod))

  def _target(self, target: Any) -> Any:
    args = None
    result = None
    auth = False

    if isinstance(target, QxHandlerFunc):
      args = target.args
      result = target.result
      auth = target.auth
      target = target.func
    elif isinstance(target, (classmethod, staticmethod)):
      target = target.__func__

    args = self.arg('args', args)
    result = self.arg('result', result)
    auth = self.arg('auth', auth)

    func = QxHandlerFunc(
      func=target,
      args=args,
      result=result,
      auth=auth,
    )

    path = self.arg([0, 'path'], getattr(target, '__qualname__', None))
    methods = self.arg(['methods', 'method'])

    if isinstance(path, str) and path:
      meta = QxHandlerMeta(
        path=path,
        methods=methods,
      )

      QxServer.get_handlers(self.arg('server')).register(QxHandler(
        func=func,
        meta=meta,
      ))

      for method in meta.methods:
        server = self.arg('server')
        server = f'[{server}]' if server else ''
        logging.info(f'{server}{method} {path} : {target}')

    return func


QxServerDecorator = QxDecorator(QxServerDecoratorHook)


def FUNC(args: Optional[Dict[str, QxAttr]] = None, result: Optional[QxAttr] = None):
  return QxServerDecorator(args=args, result=result)


def REQUEST(path: str, args: Optional[Dict[str, QxAttr]] = None, result: Optional[QxAttr] = None, auth: bool = False, methods: Optional[Union[str, List[str], Tuple[str], Set[str]]] = None):
  return QxServerDecorator(path, args=args, result=result, methods=methods, auth=auth)


def RPC(path: str, args: Optional[Dict[str, QxAttr]] = None, result: Optional[QxAttr] = None, auth: bool = False):
  return REQUEST(path, args, result, auth, _RPC_METHODS)


def HTTP(path: str, args: Optional[Dict[str, QxAttr]] = None, result: Optional[QxAttr] = None, auth: bool = False):
  return REQUEST(path, args, result, auth, _HTTP_METHODS)


def GET(path: str, args: Optional[Dict[str, QxAttr]] = None, result: Optional[QxAttr] = None, auth: bool = False):
  return REQUEST(path, args, result, auth, inspect.stack()[0][3])


def PUT(path: str, args: Optional[Dict[str, QxAttr]] = None, result: Optional[QxAttr] = None, auth: bool = False):
  return REQUEST(path, args, result, auth, inspect.stack()[0][3])


def PATCH(path: str, args: Optional[Dict[str, QxAttr]] = None, result: Optional[QxAttr] = None, auth: bool = False):
  return REQUEST(path, args, result, auth, inspect.stack()[0][3])


def DELETE(path: str, args: Optional[Dict[str, QxAttr]] = None, result: Optional[QxAttr] = None, auth: bool = False):
  return REQUEST(path, args, result, auth, inspect.stack()[0][3])


def POST(path: str, args: Optional[Dict[str, QxAttr]] = None, result: Optional[QxAttr] = None, auth: bool = False):
  return REQUEST(path, args, result, auth, inspect.stack()[0][3])


def HEAD(path: str, args: Optional[Dict[str, QxAttr]] = None, result: Optional[QxAttr] = None, auth: bool = False):
  return REQUEST(path, args, result, auth, inspect.stack()[0][3])


def OPTIONS(path: str, args: Optional[Dict[str, QxAttr]] = None, result: Optional[QxAttr] = None, auth: bool = False):
  return REQUEST(path, args, result, auth, inspect.stack()[0][3])


################################################################################



from typing import *

import os
import smtplib
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication

from quicly.logutils import QxLogging


logging = QxLogging.get_logger('QxMail')


class QxMail(object):
  def __init__(self):
    self._from_addr = None
    self._to_addrs = []
    self._cc_addrs = []
    self._bcc_addrs = []
    self._subject = None
    self._content = None
    self._attachments = []

  def __str__(self) -> str:
    return f'{self._subject} -> {self._to_addrs}'

  @property
  def from_addr(self) -> str:
    return self._from_addr

  @property
  def to_addrs(self) -> List[str]:
    return self._to_addrs

  @property
  def msg(self) -> str:
    ret = MIMEMultipart()
    ret['From'] = self._from_addr
    ret['To'] = ';'.join(self._to_addrs)
    if self._cc_addrs:
      ret['Cc'] = ';'.join(self._cc_addrs)
    if self._bcc_addrs:
      ret['Bcc'] = ';'.join(self._bcc_addrs)
    ret['Subject'] = self._subject
    if isinstance(self._content, MIMEText):
      ret.attach(self._content)
    if isinstance(self._attachments, list):
      for attach in self._attachments:
        if isinstance(attach, MIMEApplication):
          ret.attach(attach)
    return ret.as_string()

  def set_from_addr(self, from_addr: str):
    self._from_addr = from_addr
    return self  # type: QxMail

  def add_to_addrs(self, *to_addrs):
    if not isinstance(self._to_addrs, list):
      self._to_addrs = list()
    self._to_addrs += to_addrs
    return self  # type: QxMail

  def add_cc_addrs(self, *cc_addrs):
    if not isinstance(self._cc_addrs, list):
      self._cc_addrs = list()
    self._cc_addrs += cc_addrs
    return self  # type: QxMail

  def add_bcc_addrs(self, *bcc_addrs):
    if not isinstance(self._bcc_addrs, list):
      self._bcc_addrs = list()
    self._bcc_addrs += bcc_addrs
    return self  # type: QxMail

  def set_subject(self, subject: str):
    self._subject = subject
    return self  # type: QxMail

  def set_text_content(self, content: str, charset: str = 'utf-8'):
    self._content = MIMEText(content, _charset=charset)
    return self  # type: QxMail

  def set_html_content(self, content: str, charset: str = 'utf-8'):
    self._content = MIMEText(content, _subtype='html', _charset=charset)
    return self  # type: QxMail

  def add_attachment_data(self, data: bytes, filename: str = None):
    attach = MIMEApplication(data)
    attach.add_header('Content-Disposition', 'attachment', filename=filename)
    self._attachments.append(attach)
    return self  # type: QxMail

  def add_attachment(self, filename: str):
    with open(filename, 'rb') as fo:
      data = fo.read()
    return self.add_attachment_data(data=data, filename=os.path.basename(filename))


class QxMailClient(object):
  def __init__(self, username: str, password: str, host: str = None, port: int = None, ssl: bool = True):
    self._username = username
    self._password = password
    default_host = 'smtp.{}'.format(username.split('@', maxsplit=1)[1])
    default_port = smtplib.SMTP_SSL_PORT if ssl else smtplib.SMTP_PORT
    self._host = host if host else default_host
    self._port = port if port else default_port
    self._ssl = bool(ssl)
    self._smtp_client = None
    self._mails = list()

  def __enter__(self):
    self._mails = list()
    self._login()

  def __exit__(self, exc_type, exc_val, exc_tb):
    for mail in self._mails:
      self._send(mail)
    self._logout()
    self._mails = list()

  def _login(self):
    if not isinstance(self._smtp_client, smtplib.SMTP):
      smtp_client = smtplib.SMTP_SSL(host=self._host, port=self._port) if self._ssl else smtplib.SMTP(host=self._host, port=self._port)
      try:
        smtp_client.login(user=self._username, password=self._password)
      except smtplib.SMTPException:
        smtp_client.quit()
        smtp_client.close()
        smtp_client = None
      self._smtp_client = smtp_client

    if isinstance(self._smtp_client, smtplib.SMTP):
      logging.debug(f'邮箱登录成功: {self._username}')
    else:
      logging.fatal(f'邮箱登录失败: {self._username}')

    return self  # type: QxMailClient

  def _logout(self):
    if isinstance(self._smtp_client, smtplib.SMTP):
      self._smtp_client.quit()
      self._smtp_client.close()
      logging.debug(f'邮箱登出成功: {self._username}')
    self._smtp_client = None

  def _send(self, mail: QxMail):
    if isinstance(self._smtp_client, smtplib.SMTP):
      if mail.to_addrs:
        try:
          self._smtp_client.sendmail(
            from_addr=mail.from_addr if mail.from_addr else self._username,
            to_addrs=mail.to_addrs,
            msg=mail.msg,
          )
          logging.debug(f'邮件发送成功: {mail}')
        except smtplib.SMTPException as e:
          logging.error(f'邮件发送失败: {mail}')
          logging.error(e)
      else:
        logging.error(f'邮件发送失败: {mail}, 没有指定收件人地址')
    else:
      logging.error(f'邮箱未成功登录,无法发送邮件: {mail}')

  def new_mail(self) -> QxMail:
    self._mails.append(QxMail().set_from_addr(self._username))
    return self._mails[-1]



from typing import *

import os
import hashlib
from .decorator import *


class QxHash(object):
  MD5 = 'md5'
  SHA1 = 'sha1'
  SHA224 = 'sha224'
  SHA256 = 'sha256'
  SHA384 = 'sha384'
  SHA512 = 'sha512'
  BLAKE2B = 'blake2b'
  BLAKE2S = 'blake2s'
  SHA3_224 = 'sha3_224'
  SHA3_256 = 'sha3_256'
  SHA3_384 = 'sha3_384'
  SHA3_512 = 'sha3_512'
  SHAKE_128 = 'shake_128'
  SHAKE_256 = 'shake_256'

  def __init__(self, algorithm: str = SHA1):
    self._hash_o = getattr(hashlib, algorithm)
    assert callable(self._hash_o)

  @hsm
  def hash_s(self, s: Union[str, bytes]) -> str:
    if isinstance(s, str):
      s = s.encode('utf-8')
    h = self._hash_o()
    h.update(s)
    return h.hexdigest().lower()

  @hfm
  def hash_f(self, f: str) -> str:
    ret = None
    if os.path.isfile(f):
      with open(f, 'rb') as fo:
        ret = self.hash_s(fo.read())
    elif os.path.isdir(f):
      lines = []
      for item in sorted(os.listdir(f)):
        pathname = os.path.join(f, item)
        lines.append('{}:{}'.format(item, self.hash_f(pathname)))
      ret = self.hash_s('\n'.join(lines))
    return ret



from typing import *

import flask
from quicly.server import QxServerHook, QxServer, QxRequest, QxResponse
from quicly.session import QxSessionFactory


class QxFlaskServerHook(QxServerHook):
  def render_template(self, name, **kw):
    return flask.render_template(name, **kw)


class QxFlaskServerApp(object):
  def __init__(
      self,
      name: Optional[str] = None,
      hook: Optional[QxServerHook] = None,
      port: int = 8080,
      host: str = '0.0.0.0',
      debug: bool = True,
      **kw,
  ):
    self.name = name
    self.hook = hook if isinstance(hook, QxServerHook) else QxFlaskServerHook()
    self.port = port
    self.host = host
    self.debug = debug
    self.static_url_path = kw.get('static_url_path', '/static/')
    self.static_folder = kw.get('static_folder', 'static')
    self.template_folder = kw.get('template_folder', 'templates')
    self.flask_app = flask.Flask(
      f'QUICLY.{self.name}' if isinstance(self.name, str) and len(self.name) else 'QUICLY',
      static_url_path=self.static_url_path,
      static_folder=self.static_folder,
      template_folder=self.template_folder,
    )
    self.server = QxServer(self.name, self.hook)
    self.flask_app.before_request(self._handle)
    self.flask_app.after_request(self._after_request)

  def _is_static_request(self):
    return isinstance(self.static_url_path, str) and len(self.static_url_path) and flask.request.path.startswith(self.static_url_path)

  def _handle(self):
    if self._is_static_request():
      return

    method = flask.request.method.upper()
    url = flask.request.url
    headers = dict()
    cookies = dict()
    body = flask.request.data
    form = dict()
    files = list()

    for k, v in flask.request.headers.items():
      headers[k] = v

    for k, v in flask.request.cookies.items():
      cookies[k] = v

    for k, v in flask.request.form.items():
      form[k] = v

    for file in flask.request.files:
      files.append(file)

    request = QxRequest(
      method=method,
      url=url,
      headers=headers,
      cookies=cookies,
      body=body,
      form=form,
      files=files,
    )

    response = self.server.handle(request)  # type: QxResponse

    res = flask.make_response(response.data, 200 if response.code == 0 else response.code, response.headers)
    for k, v in response.cookies.items():
      if v is None:
        res.delete_cookie(k)
      else:
        res.set_cookie(k, v)
    return res

  def _after_request(self, res):
    if self._is_static_request():
      res.headers.set('Server', QxServer.get_http_header_server_name())
    return res

  @staticmethod
  def _disable_flask_logging():
    import logging
    logger = logging.getLogger('werkzeug')
    logger.setLevel(logging.ERROR)

  def run(self):
    if not self.debug:
      self._disable_flask_logging()
    self.server.init()
    self.flask_app.run(host=self.host, port=self.port, debug=self.debug)



from typing import *

import base64
import calendar
import json
from datetime import date, datetime


JSONEncoder = json.JSONEncoder
JSONDecoder = json.JSONDecoder
JSONDecodeError = json.JSONDecodeError


_DEFAULT_SMART_DATE_SUFFIX = '_str'


def _process_obj(obj: Any, smart_date: bool = True, smart_date_suffix: str = None):
    smart_date_suffix = smart_date_suffix if isinstance(smart_date_suffix, str) else _DEFAULT_SMART_DATE_SUFFIX

    if obj is None or isinstance(obj, (str, int, float, bool)):
        ret = obj
    elif isinstance(obj, bytes):
        try:
            ret = obj.decode('utf-8')
        except UnicodeDecodeError:
            ret = 'data:{}'.format(base64.b64encode(obj).decode('utf-8'))
    elif isinstance(obj, (list, tuple, set)):
        ret = list()
        for i in range(len(obj)):
            ret.append(_process_obj(obj[i], smart_date, smart_date_suffix))
    elif isinstance(obj, dict):
        ret = dict()
        for k, v in obj.items():
            k = str(k)
            ret[k] = _process_obj(v, smart_date, smart_date_suffix)
            if smart_date:
                if isinstance(v, datetime):
                    ret[f'{k}{smart_date_suffix}'] = v.strftime('%Y-%m-%d %H:%M:%S')
                elif isinstance(v, date):
                    ret[f'{k}{smart_date_suffix}'] = v.strftime('%Y-%m-%d')
    elif isinstance(obj, date):
        ret = calendar.timegm(obj.timetuple())
    else:
        ret = str(obj)

    return ret


def load(
    fp,
    *,
    cls: Optional[Type[JSONDecoder]] = None,
    object_hook: Optional[dict] = None,
    parse_float: Optional[Callable] = None,
    parse_int: Optional[Callable] = None,
    parse_constant: Optional[Callable] = None,
    object_pairs_hook: Optional[Callable] = None,
    **kw
):
    return json.load(
        fp=fp,
        cls=cls,
        object_hook=object_hook,
        parse_float=parse_float,
        parse_int=parse_int,
        parse_constant=parse_constant,
        object_pairs_hook=object_pairs_hook,
        **kw
    )


def dump(
    obj,
    fp,
    *,
    skipkeys: bool = False,
    ensure_ascii: bool = True,
    check_circular: bool = True,
    allow_nan: bool = True,
    cls: Optional[Type[JSONEncoder]] = None,
    indent: Union[None, int, str] = None,
    separators: Optional[Tuple[str, str]] = None,
    default: Optional[Callable] = None,
    sort_keys: bool = False,
    smart_date: bool = True,
    smart_date_suffix: str = None,
    **kw
):
    return json.dump(
        _process_obj(obj, smart_date, smart_date_suffix),
        fp=fp,
        skipkeys=skipkeys,
        ensure_ascii=ensure_ascii,
        check_circular=check_circular,
        allow_nan=allow_nan,
        cls=cls,
        indent=indent,
        separators=separators,
        default=default,
        sort_keys=sort_keys,
        **kw
    )


def loads(
    s: Union[str, bytes],
    cls: Optional[Type[JSONDecoder]] = None,
    object_hook: Optional[dict] = None,
    parse_float: Optional[Callable] = None,
    parse_int: Optional[Callable] = None,
    parse_constant: Optional[Callable] = None,
    object_pairs_hook: Optional[Callable] = None,
    **kw
):
    return json.loads(
        s,
        cls=cls,
        object_hook=object_hook,
        parse_float=parse_float,
        parse_int=parse_int,
        parse_constant=parse_constant,
        object_pairs_hook=object_pairs_hook,
        **kw
    )


def dumps(
    obj: Any,
    skipkeys: bool = False,
    ensure_ascii: bool = True,
    check_circular: bool = True,
    allow_nan: bool = True,
    cls: Optional[Type[JSONEncoder]] = None,
    indent: Union[None, int, str] = None,
    separators: Optional[Tuple[str, str]] = None,
    default: Optional[Callable] = None,
    sort_keys: bool = False,
    smart_date: bool = True,
    smart_date_suffix: str = None,
    **kw
):
    return json.dumps(
        _process_obj(obj, smart_date, smart_date_suffix),
        skipkeys=skipkeys,
        ensure_ascii=ensure_ascii,
        check_circular=check_circular,
        allow_nan=allow_nan,
        cls=cls,
        indent=indent,
        separators=separators,
        default=default,
        sort_keys=sort_keys,
        **kw)



import requests


class DingRobot(object):
    def __init__(self, access_token):
        self._access_token = access_token
        self._webhook_url = 'https://oapi.dingtalk.com/robot/send?access_token={}'.format(access_token)

    @property
    def access_token(self):
        return self._access_token

    @property
    def webhook_url(self):
        return self._webhook_url

    def send_message(self, data):
        assert isinstance(data, dict)
        requests.post(self._webhook_url, json=data, headers={
            'Content-Type': 'application/json',
        })


class DingMessageBuilder(object):
    def build(self):
        raise NotImplementedError()


class DingTextMessageBuilder(DingMessageBuilder):
    def __init__(self):
        self._text = []
        self._at_mobiles = []
        self._is_at_all = False

    def at(self, mobile):
        if isinstance(mobile, str):
            self._at_mobiles.append(mobile)
        elif isinstance(mobile, (list, tuple, set)):
            self._at_mobiles.append(x for x in mobile if isinstance(x, str))
        return self

    def at_all(self, is_at_all=True):
        self._is_at_all = is_at_all
        return self

    def set_text(self, text):
        self._text = [text]
        return self

    def clear_text(self):
        self._text = []
        return self

    def append_text(self, text):
        self._text.append(text)
        return self

    def append_at(self, mobile):
        self._text.append(u'@{}'.format(mobile))
        return self

    def append_newline(self):
        self._text.append(u'\n')
        return self

    def build(self):
        msg = {
            'msgtype': 'text',
            'text': {
                'content': ''.join(self._text),
            },
            'at': {
                'atMobiles': self._at_mobiles,
                'isAtAll': self._is_at_all,
            },
        }
        return msg


class DingLinkMessageBuilder(DingMessageBuilder):
    def __init__(self):
        self._title = u''
        self._text = u''
        self._url = None
        self._image_url = None

    def set_title(self, title):
        self._title = title
        return self

    def set_text(self, text):
        self._text = text
        return self

    def set_url(self, url):
        self._url = url
        return self

    def set_image_url(self, image_url):
        self._image_url = image_url
        return self

    def build(self):
        msg = {
            'msgtype': 'link',
            'link': {
                'title': self._title,
                'text': self._text,
                'messageUrl': self._url,
                'picUrl': self._image_url,
            },
        }
        return msg


class _DingMarkdownMessageBuilderBase(DingMessageBuilder):
    def __init__(self):
        self._text = []

    def set_text(self, text):
        self._text = [text]
        return self

    def clear_text(self):
        self._text = []
        return self

    def append_text(self, text):
        if isinstance(text, str) and len(text):
            self._text.append(text)
        return self

    def append_newline(self):
        self._text.append('\n')

    def ensure_newline(self):
        is_newline = False
        if len(self._text) == 0:
            is_newline = True
        elif self._text[-1].endswith('\n'):
            is_newline = True
        if not is_newline:
            self.append_newline()

    def append_h1(self, title):
        self.ensure_newline()
        self._text.append('# {}'.format(title))
        self.append_newline()
        return self

    def append_h2(self, title):
        self.ensure_newline()
        self._text.append('## {}'.format(title))
        self.append_newline()
        return self

    def append_h3(self, title):
        self.ensure_newline()
        self._text.append('### {}'.format(title))
        self.append_newline()
        return self

    def append_h4(self, title):
        self.ensure_newline()
        self._text.append('#### {}'.format(title))
        self.append_newline()
        return self

    def append_h5(self, title):
        self.ensure_newline()
        self._text.append('##### {}'.format(title))
        self.append_newline()
        return self

    def append_h6(self, title):
        self.ensure_newline()
        self._text.append('###### {}'.format(title))
        self.append_newline()
        return self

    def append_ref(self, text):
        self.ensure_newline()
        self._text.append('> {}'.format(text))
        self.append_newline()
        return self

    def append_bold(self, text):
        self._text.append('**{}**'.format(text))
        return self

    def append_italic(self, text):
        self._text.append('*{}*'.format(text))
        return self

    def append_link(self, title, url):
        if not (isinstance(title, str) and len(title)):
            title = url
        self._text.append('[{}]({})'.format(title, url))
        return self

    def append_image(self, title, url):
        if not isinstance(title, str):
            title = ''
        self._text.append('![{}]()'.format(title, url))
        return self

    def append_unordered_list(self, items):
        assert isinstance(items, (list, tuple, set))
        self.ensure_newline()
        for item in items:
            self._text.append('- {}'.format(item))
            self.append_newline()
        return self

    def append_ordered_list(self, items):
        assert isinstance(items, (list, tuple, set))
        self.ensure_newline()
        for i in range(len(items)):
            item = items[i]
            self._text.append('{}. {}'.format(i, item))
            self.append_newline()
        return self


class DingMarkdownMessageBuilder(_DingMarkdownMessageBuilderBase):
    def __init__(self):
        self._title = u''
        self._at_mobiles = []
        self._is_at_all = False
        super(DingMarkdownMessageBuilder, self).__init__()

    def at(self, mobile):
        if isinstance(mobile, str):
            self._at_mobiles.append(mobile)
        elif isinstance(mobile, (list, tuple, set)):
            self._at_mobiles.append(x for x in mobile if isinstance(x, str))
        return self

    def at_all(self, is_at_all=True):
        self._is_at_all = is_at_all
        return self

    def set_title(self, title):
        self._title = title
        return self

    def build(self):
        msg = {
            'msgtype': 'markdown',
            'markdown': {
                'title': self._title,
                'text': u''.join(self._text),
            },
            'at': {
                'atMobiles': self._at_mobiles,
                'isAtAll': self._is_at_all,
            },
        }
        return msg


class DingActionCardMessageBuilder(_DingMarkdownMessageBuilderBase):
    def __init__(self):
        self._title = u''
        self._single_title = None
        self._single_url = None
        self._btns = None
        self._btn_orientation = 0
        self._hide_avatar = False
        super(DingActionCardMessageBuilder, self).__init__()

    def set_title(self, title):
        self._title = title
        return self

    def set_single_btn(self, title, url):
        self._single_title = title
        self._single_url = url
        self._btns = None
        return self

    def append_button(self, title, url):
        self._single_title = None
        self._single_url = None
        self._btns = self._btns if isinstance(self._btns, list) else []
        self._btns.append({
            'title': title,
            'actionURL': url,
        })
        return self

    def set_btn_orientation_vertical(self):
        self._btn_orientation = 1
        return self._btn_orientation

    def set_btn_orientation_horizontal(self):
        self._btn_orientation = 0
        return self._btn_orientation

    def hide_avatar(self, hide=True):
        self._hide_avatar = hide
        return self

    def build(self):
        action_card = {
            'title': self._title,
            'text': u''.join(self._text),
            'btnOrientation': self._btn_orientation,
            'hideAvatar': self._hide_avatar,
        }
        if isinstance(self._btns, list) and len(self._btns):
            action_card['btns'] = self._btns
        else:
            action_card['singleTitle'] = self._single_title
            action_card['singleURL'] = self._single_url

        msg = {
            'msgtype': 'actionCard',
            'actionCard': action_card,
        }
        return msg


class DingFeedCardMessageBuilder(DingMessageBuilder):
    def __init__(self):
        self._links = []

    def append_link(self, title, url, image_url):
        self._links.append({
            'title': title,
            'messageURL': url,
            'picURL': image_url,
        })
        return self

    def build(self):
        msg = {
            'msgtype': 'feedCard',
            'feedCard': {
                'links': self._links[:10],
            },
        }
        return msg



from typing import *

from quicly.mongodb import QxMongoModel, QxMongoClient
from bson import ObjectId


class QxSession(object):
  def __init__(self, session_id: str):
    self._session_id = session_id

  @property
  def session_id(self) -> str:
    return self._session_id

  def __enter__(self):
    self._load()
    return self

  def __exit__(self, exc_type, exc_val, exc_tb):
    self._save()

  def _load(self):
    raise NotImplementedError()

  def _save(self):
    raise NotImplementedError()

  def get(self, name: str, default: Any = None) -> Any:
    raise NotImplementedError()

  def set(self, name: str, value: Any):
    raise NotImplementedError()

  def delete(self, name: str):
    raise NotImplementedError()

  def clear(self):
    raise NotImplementedError()


class QxMongoSession(QxSession):
  def __init__(self, model: QxMongoModel, session_id: str):
    super(QxMongoSession, self).__init__(session_id)
    self._model = model

    self._session_data = dict()
    self._session_data_set_table = dict()
    self._session_data_unset_table = dict()

    self._is_loaded = False
    self._is_created = False

  @property
  def model(self) -> QxMongoModel:
    return self._model

  def _load(self):
    self._session_data = dict()
    self._session_data_set_table = dict()
    self._session_data_unset_table = dict()

    self._is_loaded = False
    self._is_created = False

  def _save(self):
    if self._is_created:
      update_table = dict()
      if self._session_data_set_table:
        update_table['$set'] = self._session_data_set_table
      if self._session_data_unset_table:
        update_table['$unset'] = self._session_data_unset_table
      if update_table:
        self.model.update_item_by_pk(QxMongoClient.make_object_id(self.session_id), update_table)
      self._session_data_set_table.clear()
      self._session_data_unset_table.clear()
    else:
      if self._session_data:
        self._session_data['_id'] = QxMongoClient.make_object_id(self.session_id)
        self.model.insert_item(self._session_data)

  def _lazy_load(self):
    if self._is_loaded:
      return

    self._is_loaded = True
    self._session_data = self._model.find_item_by_pk(QxMongoClient.make_object_id(self.session_id))

    if self._session_data:
      self._is_created = True
    else:
      self._session_data = dict()

    for k, v in self._session_data_set_table.items():
      self._session_data[k] = v

    for k in self._session_data_unset_table.keys():
      if k in self._session_data:
        del self._session_data[k]

  def get(self, name: str, default: Any = None) -> Any:
    if name not in self._session_data:
      self._lazy_load()
    return self._session_data.get(name, default)

  def set(self, name: str, value: Any):
    self._session_data[name] = value
    self._session_data_set_table[name] = value
    if name in self._session_data_unset_table:
      del self._session_data_unset_table[name]

  def delete(self, name: str):
    if name in self._session_data:
      del self._session_data[name]
    if name in self._session_data_set_table:
      del self._session_data_set_table[name]
    self._session_data_unset_table[name] = None

  def clear(self):
    self._session_data.clear()
    self._session_data_set_table.clear()
    self._session_data_unset_table.clear()

################################################################################


class QxSessionFactory(object):
  def __init__(self):
    pass

  def new_session_id(self):
    raise NotImplementedError()

  def init_session(self):
    raise NotImplementedError()

  def load_session(self, session_id: str):
    raise NotImplementedError()


class QxMongoSessionFactory(QxSessionFactory):
  def __init__(self, db: Union[QxMongoModel, QxMongoClient]):
    super(QxMongoSessionFactory, self).__init__()
    if isinstance(db, QxMongoModel):
      self._db = db.mongo_client
      self._model = db
    else:
      self._db = db
      self._model = QxMongoModel(db, 'sessions', raw_delete=True)

  @property
  def db(self) -> QxMongoClient:
    return self._db

  @property
  def model(self) -> QxMongoModel:
    return self._model

  def new_session_id(self):
    return str(ObjectId())

  def init_session(self):
    names = self.db.collection_names(include_system_collections=False)
    if self.model.collection_name not in names:
      self.db.create_collection(self.model.collection_name)

  def load_session(self, session_id: str):
    return QxMongoSession(self.model, session_id)



import re
from typing import *

import uuid
import random
import pymongo
import pymongo.database
import pymongo.results
from bson import ObjectId
from datetime import datetime
from collections import OrderedDict


class QxMongoClient(object):
  def __init__(
      self,
      host: Union[str, pymongo.MongoClient] = '127.0.0.1',
      port: int = 27017,
      username: Optional[str] = None,
      password: Optional[str] = None,
      database: str = 'admin',
      current_time: Optional[datetime] = None
  ):
    if isinstance(host, str):
      if '://' in host:
        py_mongo_client = QxMongoClient.create_py_mongo_client_from_url(host)
      else:
        py_mongo_client = QxMongoClient.create_py_mongo_client(host, port, username, password)
    else:
      assert isinstance(host, pymongo.MongoClient)
      py_mongo_client = host

    self._mongo_client = py_mongo_client
    self._mongo_db = py_mongo_client.get_database(database)
    self._current_time = current_time

  @staticmethod
  def create_py_mongo_client(host='127.0.0.1', port=27017, username=None, password=None):
    return pymongo.MongoClient(
      host=host,
      port=port,
      username=username,
      password=password,
    )

  @staticmethod
  def create_py_mongo_client_from_url(url='mongodb://127.0.0.1:27017/admin'):
    return pymongo.MongoClient(url)

  @property
  def mongo_client(self) -> pymongo.MongoClient:
    return self._mongo_client

  @property
  def db(self) -> pymongo.database:
    return self._mongo_db

  @property
  def mongo_db(self):
    return self._mongo_db

  @property
  def current_time(self) -> datetime:
    return self._current_time if isinstance(self._current_time, datetime) else datetime.now()

  @current_time.setter
  def current_time(self, v):
    self._current_time = v

  @staticmethod
  def _preprocess_sort(sort: Optional[Union[Tuple, List, OrderedDict]]):
    items = []
    if isinstance(sort, (tuple, list)):
      if len(sort) == 2 and isinstance(sort[0], str) and sort[1] in [1, -1]:
        items = [tuple(sort)]
      else:
        for item in sort:
          if isinstance(item, (tuple, list)) and len(item) == 2 and isinstance(item[0], str) and item[1] in [1, -1]:
            items.append(tuple(item))
    elif isinstance(sort, OrderedDict):
      for (k, v) in sort.items():
        if isinstance(k, str) and v in [1, -1]:
          items.append((k, v))

    items = items if len(items) else None
    return items

  @staticmethod
  def is_object_id(object_id: Union[ObjectId, str, uuid.UUID], strict: bool = False):
    ret = False
    if isinstance(object_id, ObjectId) or (isinstance(object_id, str) and re.match(r'[0-9a-f]{24}', object_id)):
      ret = True
    elif not strict and (isinstance(str, uuid.UUID) or (isinstance(object_id, str) and re.match(r'[0-9a-f]{24}', object_id.replace('-', '')))):
      ret = True
    return ret

  @staticmethod
  def make_object_id(object_id: Union[ObjectId, str, uuid.UUID]):
    if not isinstance(object_id, ObjectId):
      object_id = ObjectId(str(object_id).replace('-', ''))
    return object_id

  def get_collection(self, name: str) -> pymongo.collection.Collection:
    return self.db.get_collection(name)

  def aggregate(self, collection_name: str, pipeline: Union[Dict, List[Dict]]) -> List[Dict]:
    items = []
    cursor = self.get_collection(collection_name).aggregate(pipeline)
    for item in cursor:
      items.append(item)
    return items

  def aggregate_one(self, collection_name: str, pipeline: Dict) -> Optional[Dict]:
    cursor = self.get_collection(collection_name).aggregate(pipeline)
    for item in cursor:
      return item
    return None

  def count(self, collection_name: str, filter: Optional[Dict] = None) -> int:
    return self.get_collection(collection_name).count_documents(filter)

  def exists(self, collection_name: str, filter: Optional[Dict] = None) -> bool:
    return bool(self.get_collection(collection_name).find_one(filter))

  def find_one(self, collection_name: str, filter: Optional[Dict] = None, sort: Optional[Union[Tuple, List, OrderedDict]] = None, projection: Optional[Dict] = None, fn_process_item: Optional[Callable] = None) -> Optional[Dict]:
    sort = self._preprocess_sort(sort)

    item = self.get_collection(collection_name).find_one(filter, sort=sort, projection=projection)
    if callable(fn_process_item) and item:
      item = fn_process_item(item)

    return item

  def find_many(self, collection_name: str, filter: Optional[Dict] = None, sort: Optional[Union[Tuple, List, OrderedDict]] = None, skip: Optional[int] = None, limit: Optional[int] = None, projection: Optional[Dict] = None, fn_process_item: Optional[Callable] = None) -> List[Dict]:
    sort = self._preprocess_sort(sort)

    items = []
    if (isinstance(limit, int) and limit > 0) or limit is None:
      kw = dict()
      if sort:
        kw['sort'] = sort
      if isinstance(skip, int):
        kw['skip'] = skip
      if isinstance(limit, int):
        kw['limit'] = limit
      if projection:
        kw['projection'] = projection
      cursor = self.get_collection(collection_name).find(filter, **kw)
      if callable(fn_process_item):
        items = [fn_process_item(item) for item in cursor]
      else:
        items = [item for item in cursor]

    return items

  def find_one_random(self, collection_name: str, filter: Optional[Dict] = None, projection: Optional[Dict] = None, fn_process_item: Optional[Callable] = None) -> Optional[Dict]:
    kw = dict()
    if projection:
      kw['projection'] = projection

    cursor = self.get_collection(collection_name).find(filter, **kw)

    item = None

    total = cursor.count()
    if total > 0:
      index = random.randint(0, max(total, 999999)) % total
      cursor.skip(index)
      item = cursor.next()

      if callable(fn_process_item) and item:
        item = fn_process_item(item)

    return item

  def insert_one(self, collection_name: str, document: Dict) -> pymongo.results.InsertOneResult:
    return self.get_collection(collection_name).insert_one(document)

  def insert_many(self, collection_name: str, documents: List[Dict]) -> pymongo.results.InsertManyResult:
    return self.get_collection(collection_name).insert_many(documents)

  def delete_one(self, collection_name: str, filter: Optional[Dict]) -> pymongo.results.DeleteResult:
    return self.get_collection(collection_name).delete_one(filter)

  def delete_many(self, collection_name: str, filter: Optional[Dict]) -> pymongo.results.DeleteResult:
    return self.get_collection(collection_name).delete_many(filter)

  def update_one(self, collection_name: str, filter: Optional[Dict], update: Dict, upsert: bool = False) -> pymongo.results.UpdateResult:
    return self.get_collection(collection_name).update_one(filter, update, upsert)

  def update_one_and_return_it(self, collection_name: str, filter: Optional[Dict], update: Dict, upsert: bool = False) -> Optional[Dict]:
    result = self.update_one(collection_name, filter, update, upsert)
    if upsert and result.upserted_id:
      return self.find_one(collection_name, {
        '_id': result.upserted_id
      })
    return self.find_one(collection_name, filter)

  def update_many(self, collection_name: str, filter: Optional[Dict], update: Dict, upsert: bool = False) -> pymongo.results.UpdateResult:
    return self.get_collection(collection_name).update_many(filter, update, upsert)

  def find_one_and_delete(self, collection_name: str, filter:  Optional[Dict], projection: Optional[Dict] = None, sort: Optional[Union[Tuple, List, OrderedDict]] = None, fn_process_item: Optional[Callable] = None) -> Optional[Dict]:
    sort = self._preprocess_sort(sort)

    item = self.get_collection(collection_name).find_one_and_delete(filter, projection, sort)
    if callable(fn_process_item) and item:
      item = fn_process_item(item)
    return item

  def find_one_and_replace(self, collection_name: str, filter:  Optional[Dict], replacement: Dict, projection: Optional[Dict] = None, sort: Optional[Union[Tuple, List, OrderedDict]] = None, upsert: bool = False, return_document: bool = False, fn_process_item: Optional[Callable] = None) -> Optional[Dict]:
    sort = self._preprocess_sort(sort)

    item = self.db.get_collection(collection_name).find_one_and_replace(filter, replacement, projection, sort, upsert, return_document)
    if callable(fn_process_item) and item:
      item = fn_process_item(item)
    return item

  def find_one_and_update(self, collection_name: str, filter:  Optional[Dict], update: Dict, projection: Optional[Dict] = None, sort: Optional[Union[Tuple, List, OrderedDict]] = None, upsert: bool = False, return_document: bool = False, fn_process_item: Optional[Callable] = None) -> Optional[Dict]:
    sort = self._preprocess_sort(sort)

    item = self.db.get_collection(collection_name).find_one_and_update(filter, update, projection, sort, upsert, return_document)
    if callable(fn_process_item) and item:
      item = fn_process_item(item)
    return item

  def find_many_with_page_info(self, collection_name: str, filter=None, sort=None, skip=None, limit=None, projection=None, page_info=None, fn_process_item=None, fn_process_sort=None) -> Tuple[List[Dict], Dict]:
    filter = dict() if filter is None else filter
    sort = self._preprocess_sort(sort)

    page_info = page_info if isinstance(page_info, dict) else dict()

    total = page_info.get('total', None)
    page_num = page_info.get('page_num', None)
    next_index = page_info.get('next_index', 0)
    cut_off_time = page_info.get('cut_off_time', self.current_time)

    filter = filter.copy()
    if '_id' in filter:
      filter['_id'] = {
        '$and': [
          {'$lte': ObjectId.from_datetime(cut_off_time)},
          filter['_id'],
        ]
      }
    else:
      filter['_id'] = {'$lte': ObjectId.from_datetime(cut_off_time)}

    if total is None:
      total = self.count(collection_name, filter=filter)

    if sort and callable(fn_process_sort):
      sort_t = fn_process_sort(sort)
    else:
      sort_t = sort

    if not isinstance(skip, int):
      skip = 0

    items = self.find_many(collection_name, filter=filter, sort=sort_t, skip=next_index + skip, limit=limit, projection=projection)

    page_num = 0 if page_num is None else page_num + 1
    from_index = next_index + skip
    next_index = next_index + skip + len(items)
    has_more = next_index < total

    page_info = dict(
      total=total,
      limit=limit,
      page_num=page_num,
      from_index=from_index,
      next_index=next_index,
      cut_off_time=cut_off_time,
      has_more=has_more,
    )

    if callable(fn_process_item) and items:
      for i in range(len(items)):
        items[i] = fn_process_item(items[i])

    return items, page_info

  def collection_names(self, include_system_collections: bool = True) -> List[str]:
    return self.db.collection_names(include_system_collections)

  def create_collection(self, name: str) -> pymongo.collection.Collection:
    return self.db.create_collection(name)

  def exists_collection(self, name: str) -> bool:
    return name in self.collection_names(False)

  def create_index(self, collection_name: str, index_name: str, keys: List[Dict], **kw):
    return self.get_collection(collection_name).create_index(keys, name=index_name, **kw)

  def delete_index(self, collection_name: str, index_name: str):
    return self.get_collection(collection_name).drop_index(index_or_name=index_name)

################################################################################


class QxMongoModel(object):
  def __init__(self, mongo_client: QxMongoClient, collection_name: str, pk_name: str = '_id', ex_field_names: Optional[Dict] = None, raw_delete: bool = True):
    self._mongo_client = mongo_client
    self._collection_name = collection_name
    self._pk_name = pk_name
    self._ex_field_names = QxMongoModel._process_ex_field_names(ex_field_names)
    self._raw_delete = raw_delete

  def _make_pk_value(self, pk_value: Any) -> Any:
    if self._pk_name == '_id' and QxMongoClient.is_object_id(pk_value):
      pk_value = QxMongoClient.make_object_id(pk_value)
    return pk_value

  @staticmethod
  def _process_ex_field_names(d: Optional[Dict]) -> Dict:
    ret = d if isinstance(d, dict) else dict()
    ret.setdefault('created_time', 'created_time')
    ret.setdefault('updated_time', 'updated_time')
    ret.setdefault('is_deleted', '__is_deleted__')
    return ret

  @property
  def mongo_client(self) -> QxMongoClient:
    return self._mongo_client

  @property
  def collection_name(self) -> str:
    return self._collection_name

  @property
  def ex_field_created_time(self) -> str:
    return self._ex_field_names.get('created_time', 'created_time')

  @property
  def ex_field_updated_time(self) -> str:
    return self._ex_field_names.get('updated_time', 'updated_time')

  @property
  def ex_field_is_deleted(self) -> str:
    return self._ex_field_names.get('is_deleted', '__is_deleted__')

  @property
  def pk_name(self) -> str:
    return self._pk_name

  @property
  def raw_delete(self) -> bool:
    return self._raw_delete

  @staticmethod
  def delete_dict_k_safe(d: Dict, k: str) -> Any:
    v = None
    if k in d:
      v = d[k]
      del d[k]
    return v

  @staticmethod
  def process_filter(filter: Optional[Dict]) -> Optional[Dict]:
    if not isinstance(filter, dict):
      filter = dict()
    return filter

  def process_filter_for_not_deleted(self, filter: Optional[Dict]) -> Dict:
    filter = self.process_filter(filter)
    if self.raw_delete:
      ret = filter
    else:
      if filter:
        ret = {
          '$and': [
            {
              self.ex_field_is_deleted: {'$ne': True},
            },
            filter,
          ]
        }
      else:
        ret = {
          self.ex_field_is_deleted: {'$ne': True},
        }
    return ret

  def insert_item(self, item: Dict) -> Dict:
    self.delete_dict_k_safe(item, self.ex_field_is_deleted)

    if '_id' in item and QxMongoClient.is_object_id(item['_id']):
      item['_id'] = QxMongoClient.make_object_id(item['_id'])
    item[self.ex_field_created_time] = self.mongo_client.current_time
    item[self.ex_field_updated_time] = self.mongo_client.current_time

    self.mongo_client.insert_one(self.collection_name, item)

    return item

  def delete_item(self, filter: Optional[Dict]):
    filter = self.process_filter_for_not_deleted(filter)
    if self.raw_delete:
      self.mongo_client.delete_one(self.collection_name, filter)
    else:
      self.mongo_client.update_one(self.collection_name, filter, {
        '$set': {
          self.ex_field_updated_time: self.mongo_client.current_time,
          self.ex_field_is_deleted: True,
        }
      })

  def delete_item_by_property(self, property_name: str, property_value: Any):
    self.delete_item({
      property_name: property_value,
    })

  def delete_item_by_pk(self, pk_value: Any):
    self.delete_item_by_property(property_name=self.pk_name, property_value=pk_value)

  def delete_items(self, filter: Optional[Dict]):
    filter = self.process_filter_for_not_deleted(filter)
    if self.raw_delete:
      self.mongo_client.delete_many(self.collection_name, filter)
    else:
      self.mongo_client.update_many(self.collection_name, filter, {
        '$set': {
          self.ex_field_updated_time: self.mongo_client.current_time,
          self.ex_field_is_deleted: True,
        }
      })

  def update_item(self, filter: Optional[Dict], update: Dict, upsert: bool = False):
    filter = self.process_filter_for_not_deleted(filter)

    set_table = update.get('$set', {})
    set_table[self.ex_field_updated_time] = self.mongo_client.current_time
    update['$set'] = set_table

    self.mongo_client.update_one(self.collection_name, filter, update, upsert)

  def update_items(self, filter: Optional[Dict], update: Dict, upsert: bool = False):
    filter = self.process_filter_for_not_deleted(filter)

    set_table = update.get('$set', {})
    set_table[self.ex_field_updated_time] = self.mongo_client.current_time
    update['$set'] = set_table

    self.mongo_client.update_many(self.collection_name, filter, update, upsert)

  def update_item_by_property(self, property_name: str, property_value: Any, update: Dict, upsert: bool = False):
    self.update_item({
      property_name: property_value,
    }, update, upsert)

  def update_item_by_pk(self, pk_value: Any, update: Dict, upsert: bool = False):
    self.update_item_by_property(self.pk_name, pk_value, update, upsert)

  def set_item_property(self, pk_value: Any, property_name: str, property_value: Any):
    self.update_item_by_pk(pk_value, {
      '$set': {
        property_name: property_value,
      },
    })

  def inc_item_property(self, pk_value: Any, property_name: str, property_value: Any):
    self.update_item_by_pk(pk_value, {
      '$inc': {
        property_name: property_value,
      },
    })

  def push_item_property(self, pk_value: Any, property_name: str, property_value: Any):
    self.update_item_by_pk(pk_value, {
      '$push': {
        property_name: property_value,
      },
    })

  def pull_item_property(self, pk_value: Any, property_name: str, property_value: Any):
    self.update_item_by_pk(pk_value, {
      '$pull': {
        property_name: property_value,
      },
    })

  def update_item_properties(self, pk_value: Any, set_properties: Optional[Dict] = None, inc_properties: Optional[Dict] = None, push_properties: Optional[Dict] = None, pull_properties: Optional[Dict] = None):
    update = dict()
    if isinstance(set_properties, dict) and len(set_properties):
      update['$set'] = set_properties
    if isinstance(inc_properties, dict) and len(inc_properties):
      update['$inc'] = inc_properties
    if isinstance(push_properties, dict) and len(push_properties):
      update['$push'] = push_properties
    if isinstance(pull_properties, dict) and len(pull_properties):
      update['$pull'] = pull_properties
    if len(update):
      self.update_item_by_pk(pk_value, update)

  def set_item_properties(self, pk_value: Any, **kw):
    self.update_item_properties(pk_value, set_properties=kw)

  def inc_item_properties(self, pk_value: Any, **kw):
    self.update_item_properties(pk_value, inc_properties=kw)

  def push_item_properties(self, pk_value: Any, **kw):
    self.update_item_properties(pk_value, push_properties=kw)

  def pull_item_properties(self, pk_value: Any, **kw):
    self.update_item_properties(pk_value, pull_properties=kw)

  def find_item(self, filter: Optional[Dict], default: Optional[Dict] = None) -> Optional[Dict]:
    filter = self.process_filter_for_not_deleted(filter)
    item = self.mongo_client.find_one(self.collection_name, filter)
    item = default if item is None else item
    return item

  def find_item_random(self, filter: Optional[Dict], default: Optional[Dict] = None):
    filter = self.process_filter_for_not_deleted(filter)
    item = self.mongo_client.find_one_random(self.collection_name, filter)
    item = default if item is None else item
    return item

  def find_item_by_property(self, property_name: str, property_value: Any):
    return self.find_item({
      property_name: property_value,
    })

  def find_item_by_pk(self, pk_value: Any):
    return self.find_item_by_property(self.pk_name, pk_value)

  def find_items_with_page_info(self, filter: Optional[Dict], sort: Union[Tuple, List, OrderedDict] = None, limit: int = None, page_info: Optional[Dict] = None):
    filter = self.process_filter_for_not_deleted(filter)
    items, page_info = self.mongo_client.find_many_with_page_info(
      self.collection_name,
      filter=filter,
      sort=sort,
      limit=limit,
      page_info=page_info,
    )
    return items, page_info

  def find_items(self, filter: Optional[Dict], sort: Union[Tuple, List, OrderedDict] = None):
    items, page_info = self.find_items_with_page_info(filter, sort)
    return items

  def find_items_by_property_as_table(self, property_name: str, property_values: Union[List, Tuple, Set]) -> Dict:
    assert isinstance(property_values, (set, list, tuple))
    items = self.find_items({
      property_name: {'$in': list(property_values)},
    })
    table = dict()
    for item in items:
      if property_name in item:
        table[item[property_name]] = item
    return table

  def find_items_by_property_as_order(self, property_name: str, property_values: Union[List, Tuple, Set]) -> List[Dict]:
    table = self.find_items_by_property_as_table(property_name, property_values)
    items = [table[x] for x in property_values if x in table]
    return items

  def find_items_by_pk_as_table(self, pk_values: Union[List, Tuple, Set]):
    return self.find_items_by_property_as_table(self.pk_name, pk_values)

  def find_items_by_pk_as_order(self, pk_values: Union[List, Tuple, Set]):
    table = self.find_items_by_pk_as_table(pk_values=pk_values)
    items = [table[x] for x in pk_values if x in table]
    return items

  def is_exists(self, filter: Optional[Dict]) -> bool:
    return bool(self.find_item(filter))

  def is_exists_by_property(self, property_name: str, property_value: Any) -> bool:
    return bool(self.find_item_by_property(property_name, property_value))

  def is_exists_by_pk(self, pk_value: Any) -> bool:
    return bool(self.find_item_by_pk(pk_value))

  def count(self, filter: Optional[Dict]) -> int:
    return self.mongo_client.count(self.collection_name, filter)



import logging


class QxLogging(object):
  DEFAULT_LEVEL = 'DEBUG'
  DEFAULT_FORMAT = '[%(asctime)s] [%(name)s] %(message)s'

  def __init__(self, name: str = None):
    self._logger = self.get_logger(name)

  @property
  def logger(self) -> logging.Logger:
    return self._logger

  @staticmethod
  def get_logger(name: str):
    return logging.getLogger(name)

  @staticmethod
  def disable(name: str):
    QxLogging.get_logger(name).disabled = True


logging.basicConfig(level=QxLogging.DEFAULT_LEVEL, format=QxLogging.DEFAULT_FORMAT)



from typing import *

from quicly.value import *
from quicly.limits import *


class QxxAttr(object):
  def __init__(self):
    super(QxxAttr, self).__init__()

  def check(self, v: Any) -> bool:
    raise NotImplementedError()

  def value(self, v: Any, default: Any = None) -> Any:
    raise NotImplementedError()


class QxxAny(QxxAttr):
  def check(self, v: Any) -> bool:
    return True

  def value(self, v: Any, default: Any = None) -> Any:
    return v


class QxxOptional(QxxAttr):
  def __init__(self, element: QxxAttr, default: Any = None):
    super(QxxOptional, self).__init__()
    assert isinstance(element, QxxAttr)
    self._element = element
    self._default = default

  @property
  def element(self) -> QxxAttr:
    return self._element

  @property
  def default(self) -> Any:
    return self._default

  def check(self, v: Any) -> bool:
    v = QxValue.create(v)
    return v.is_none() or self.element.check(v)

  def value(self, v: Any, default: Any = None) -> Any:
    return self.element.value(v, default=default)


class QxxUnion(QxxAttr):
  def __init__(self, elements: List[QxxAttr]):
    super(QxxUnion, self).__init__()
    self._elements = elements

  @property
  def elements(self) -> List[QxxAttr]:
    return self._elements

  def match(self, v: Any) -> QxxAttr:
    ret = None

    for elem in self.elements:
      if elem.check(v):
        ret = elem
        break

    return ret

  def check(self, v: Any) -> bool:
    elem = self.match(v)

    return isinstance(elem, QxxAttr)

  def value(self, v: Any, default: Any = None) -> Any:
    elem = self.match(v)

    if isinstance(elem, QxxAttr):
      ret = elem.value(v, default)
    else:
      ret = default

    return ret


class QxxBool(QxxAttr):
  def check(self, v: Any) -> bool:
    return QxValue.create(v).is_bool()

  def value(self, v: Any, default: Optional[bool] = None) -> bool:
    return QxValue.create(v).to_bool(default=default)


class QxxInt(QxxAttr):
  LIMITS_MAX = sys.maxsize
  LIMITS_MIN = -LIMITS_MAX - 1

  def __init__(self, limits: Optional[Tuple[Optional[int], Optional[int]]] = None):
    super(QxxInt, self).__init__()
    limits = list(limits) if isinstance(limits, (tuple, list)) and len(limits) == 2 else [self.LIMITS_MIN, self.LIMITS_MAX]
    limits[0] = max(limits[0], self.LIMITS_MIN) if isinstance(limits[0], int) else self.LIMITS_MIN
    limits[1] = min(limits[1], self.LIMITS_MAX) if isinstance(limits[1], int) else self.LIMITS_MAX
    self._limits = (limits[0], limits[1])

  @property
  def limits(self) -> Tuple[int, int]:
    return self._limits

  @property
  def limits_min(self) -> int:
    return self._limits[0]

  @property
  def limits_max(self) -> int:
    return self._limits[1]

  def check(self, v: Any) -> bool:
    v = QxValue.create(v)
    return v.is_int() and self.limits_min <= v.int_value <= self.limits_max

  def value(self, v: Any, default: Optional[int] = None) -> int:
    return QxValue.create(v).to_int(default=default)


class QxxInt32(QxxInt):
  LIMITS_MAX = INT32_MAX
  LIMITS_MIN = INT32_MIN


class QxxInt64(QxxInt):
  LIMITS_MAX = INT64_MAX
  LIMITS_MIN = INT64_MIN


class QxxUint32(QxxInt):
  LIMITS_MAX = UINT32_MAX
  LIMITS_MIN = UINT32_MIN


class QxxUint64(QxxInt):
  LIMITS_MAX = UINT64_MAX
  LIMITS_MIN = UINT64_MIN


class QxxFloat(QxxAttr):
  LIMITS_MAX = float('inf')
  LIMITS_MIN = float('-inf')
  EPSILON = FLOAT64_EPSILON

  def __init__(self, limits: Optional[Tuple[Optional[float], Optional[float]]] = None):
    super(QxxFloat, self).__init__()
    limits = list(limits) if isinstance(limits, (tuple, list)) and len(limits) == 2 else [self.LIMITS_MIN, self.LIMITS_MAX]
    limits[0] = max(limits[0], self.LIMITS_MIN - self.EPSILON) if isinstance(limits[0], float) else self.LIMITS_MIN
    limits[1] = min(limits[1], self.LIMITS_MAX + self.EPSILON) if isinstance(limits[1], float) else self.LIMITS_MAX
    self._limits = (limits[0], limits[1])

  @property
  def limits(self) -> Tuple[float, float]:
    return self._limits

  @property
  def limits_min(self) -> float:
    return self._limits[0]

  @property
  def limits_max(self) -> float:
    return self._limits[1]

  def check(self, v: Any) -> bool:
    v = QxValue.create(v)
    return v.is_float() and self.limits_min - self.EPSILON <= v.int_value <= self.limits_max + self.EPSILON

  def value(self, v: Any, default: Optional[float] = None) -> float:
    return QxValue.create(v).to_float(default=default)


class QxxFloat32(QxxFloat):
  LIMITS_MAX = FLOAT32_MAX
  LIMITS_MIN = FLOAT32_MIN
  EPSILON = FLOAT32_EPSILON


class QxxFloat64(QxxFloat):
  LIMITS_MAX = FLOAT64_MAX
  LIMITS_MIN = FLOAT64_MIN
  EPSILON = FLOAT64_EPSILON


class QxxStr(QxxAttr):
  def check(self, v: Any) -> bool:
    return QxValue.create(v).is_str()

  def value(self, v: Any, default: Optional[str] = None) -> str:
    return QxValue.create(v).to_str(default=default)


class QxxList(QxxAttr):
  def __init__(self, element: Optional[QxxAttr] = None):
    super(QxxList, self).__init__()
    self._element = element if isinstance(element, QxxAttr) else None

  @property
  def element(self) -> QxxAttr:
    return self._element

  def check(self, v: Any) -> bool:
    v = QxValue.create(v)

    if v.is_list():
      ret = True
      if isinstance(self.element, QxxAttr):
        for i in v.list_value:
          if not self.element.check(i):
            ret = False
            break
    else:
      ret = False

    return ret

  def value(self, v: Any, default: Optional[List[Any]] = None) -> List[Any]:
    return QxValue.create(v).to_list(default=default)


class QxxDict(QxxAttr):
  def __init__(self, element: Optional[QxxAttr] = None):
    super(QxxDict, self).__init__()
    self._element = element if isinstance(element, QxxAttr) else None

  @property
  def element(self):
    return self._element

  def check(self, v: Any) -> bool:
    v = QxValue.create(v)

    if v.is_dict():
      ret = True
      if isinstance(self.element, QxxAttr):
        for kk, vv in v.dict_value.items():
          if not self.element.check(vv):
            ret = False
            break
    else:
      ret = False

    return ret

  def value(self, v: Any, default: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
    return QxValue.create(v).to_dict(default)



from typing import *

from quicly.conf import QxGlobalConf
from quicly import jsonutils as json
from quicly.value import QxValue


class QxResult(Exception):
  CODE_0_OK = 0
  CODE_400_BAD_REQUEST = 400
  CODE_401_UNAUTHORIZED = 401
  CODE_403_FORBIDDEN = 403
  CODE_404_NOT_FOUND = 404
  CODE_405_METHOD_NOT_ALLOWED = 405
  CODE_500_INTERNAL_SERVER_ERROR = 500
  CODE_501_NOT_IMPLEMENTED = 501

  def __init__(self, code: int, reason: str, data: Any, debug: dict):
    super(QxResult, self).__init__()
    self._code = self._process_code(code)
    self._reason = self._process_reason(reason, self._code)
    self._message = self._process_message(self._reason)
    self._data = self._process_data(data)
    self._debug = self._process_debug(debug)

  @classmethod
  def mk_code(cls, reason: Optional[str] = None) -> int:
    conf = QxGlobalConf()
    return conf.code.get(reason, cls.CODE_500_INTERNAL_SERVER_ERROR)

  @classmethod
  def mk_reason(cls, code: Optional[int] = None) -> str:
    conf = QxGlobalConf()
    return conf.code.get(code, conf.code.get(cls.CODE_500_INTERNAL_SERVER_ERROR))

  @classmethod
  def mk_message(cls, reason: Optional[str] = None, code: Optional[int] = None):
    conf = QxGlobalConf()
    reason = reason if isinstance(reason, str) and reason else cls.mk_reason(code)
    return conf.i18n.get(reason, reason)

  @classmethod
  def _process_code(cls, code: Optional[int] = None) -> int:
    return QxValue(code).to_int(cls.CODE_500_INTERNAL_SERVER_ERROR)

  @classmethod
  def _process_reason(cls, reason: Optional[str] = None, code: Optional[int] = None) -> str:
    if isinstance(reason, str) and reason:
      ret = reason
    else:
      ret = cls.mk_reason(code)
    return ret

  @classmethod
  def _process_message(cls, reason: Optional[str] = None, code: Optional[int] = None) -> str:
    return cls.mk_message(reason, code)

  @staticmethod
  def _process_data(data: Any) -> Any:
    return data

  @staticmethod
  def _process_debug(debug: dict) -> dict:
    return debug

  @property
  def code(self) -> int:
    return self._code

  @property
  def reason(self) -> str:
    return self._reason

  @property
  def message(self) -> str:
    return self._message

  @property
  def data(self) -> Any:
    return self._data

  @property
  def debug(self) -> Union[dict, None]:
    return self._debug

  def __bool__(self):
    return self._code == 0

  def to_dict(self):
    ret = {
      'code': self._code,
      'reason': self._reason,
      'message': self._message,
      'data': self._data,
    }
    if QxGlobalConf().is_debug:
      ret['debug'] = self._debug
    return ret

  def __str__(self):
    return json.dumps(self.to_dict(), indent=2, ensure_ascii=False)

  def __call__(self):
    raise self


class QxError(QxResult):
  def __init__(self, code: int = None, reason: str = None, data: Any = None, debug: dict = None):
    super(QxError, self).__init__(
      code=code if code != self.CODE_0_OK else self.CODE_500_INTERNAL_SERVER_ERROR,
      reason=reason,
      data=data,
      debug=debug,
    )


class QxFinish(QxResult):
  def __init__(self, data: Any = None, reason: str = None, debug: dict = None):
    super(QxFinish, self).__init__(
      code=0,
      reason=reason,
      data=data,
      debug=debug,
    )

################################################################################


class QxHttpResult(QxFinish):
  def __init__(self, status_code: int = 200, content: Union[str, bytes, dict] = '', headers: Optional[dict] = None, cookies: Optional[dict] = None):
    super(QxHttpResult, self).__init__(data={
      'status_code': status_code,
      'content': content,
      'headers': headers if isinstance(headers, dict) else dict(),
      'cookies': cookies if isinstance(cookies, dict) else dict(),
    })

  @property
  def code(self):
    return self.data['status_code']

  @property
  def content(self):
    return self.data['content']

  @property
  def headers(self):
    return self.data['headers']

  @property
  def cookies(self):
    return self.data['cookies']

  def set_header(self, name: str, value: str):
    self.data['headers'][name] = value
    return self

  def set_cookie(self, name: str, value: Optional[str]):
    self.data['cookies'][name] = value
    return self


class QxStatic(QxHttpResult):
  def __init__(self, path: str):
    super(QxStatic, self).__init__(content=path)


class QxRedirect(QxHttpResult):
  def __init__(self, url):
    super(QxRedirect, self).__init__(
      status_code=302,
      headers={
        'Location': url,
      }
    )


class QxHtml(QxHttpResult):
  def __init__(self, html, content_type: str = 'text/html', status_code: int = 200):
    super(QxHtml, self).__init__(
      status_code=status_code,
      content=html,
      headers={
        'Content-Type': content_type,
      },
    )


class QxTemplate(QxHttpResult):
  def __init__(self, name: str, args: Any = None, status_code: int = 200):
    super(QxTemplate, self).__init__(
      status_code=status_code,
      content={
        'name': name,
        'args': args,
      },
    )



from typing import *

from quicly.value import QxValue

from collections import OrderedDict


class QxScriptContext(object):
  def __init__(self, parent: None):
    assert parent is None or isinstance(parent, QxScriptContext)
    self.parent = parent
    self.vars = OrderedDict()

  def set(self, name: str, value: Any) -> NoReturn:
    self.vars[name] = QxValue(value)

  def has(self, name: str, local: bool = False) -> bool:
    if name in self.vars:
      ret = True
    elif not local and self.parent:
      ret = self.parent.has(name, local)
    else:
      ret = False
    return ret

  def get(self, name: str, default: Any = None, local: bool = False) -> QxValue:
    if name in self.vars:
      ret = self.vars.get(name, default)
    elif not local and self.parent:
      ret = self.parent.get(name, default)
    else:
      ret = default
    return QxValue(ret)

  def delete(self, name: str, local: bool = True) -> bool:
    if name in self.vars:
      del self.vars[name]
      ret = True
    elif not local and self.parent:
      ret = self.parent.delete(name, local)
    else:
      ret = False
    return ret



from typing import *

from quicly.patterns import PATTERN_UUID

from bson import ObjectId
from uuid import UUID, uuid4


class QxUuidUtils(object):
  @staticmethod
  def mk_uuid(v: Any = None, sp: Optional[str] = '-') -> str:
    if isinstance(v, ObjectId):
      ret = str(v)
    elif isinstance(v, UUID):
      ret = str(v).replace('-', '')
    elif isinstance(v, str) and PATTERN_UUID.fullmatch(v):
      ret = ''.join(PATTERN_UUID.findall(v))
    else:
      ret = str(ObjectId())

    if isinstance(sp, str) and len(sp):
      ret = sp.join([ret[0:8], ret[8:12],  ret[12:16], ret[16:20], ret[20:32]])

    return ret



from typing import *

from quicly.cache import QxCache

import threading


QX_FUNCTION_TYPE = type(lambda: None)


class QxTyping(object):
  def check_typing(self, v: Any) -> bool:
    raise NotImplementedError()

  def assert_typing(self, v: Any) -> NoReturn:
    assert self.check_typing(v)


class QxCustomTyping(QxTyping):
  def __init__(self, t: Type):
    self.t = t

  def check_typing(self, v: Any) -> bool:
    ret = False

    if self.t is None and v is None:
      ret = True
    elif isinstance(self.t, type) and isinstance(v, self.t):
      ret = True
    elif isinstance(self.t, QxTyping):
      ret = self.t.check_typing(v)

    return ret


class _QxCoreTyping(QxTyping):
  _lock = threading.Lock()
  _inst = None

  @classmethod
  def __new__(cls, *a, **kw):
    if not isinstance(cls._inst, cls):
      with cls._lock:
        if not isinstance(cls._inst, cls):
          cls._inst = super().__new__(cls)
    return cls._inst

  def check_typing(self, v: Any) -> bool:
    raise NotImplementedError()


class QxAnyTyping(_QxCoreTyping):
  def check_typing(self, v: Any) -> bool:
    return True


class QxNoneTyping(_QxCoreTyping):
  def check_typing(self, v: Any) -> bool:
    return v is None


class QxBoolTyping(_QxCoreTyping):
  def check_typing(self, v: Any) -> bool:
    return isinstance(v, bool)


class QxIntTyping(_QxCoreTyping):
  def check_typing(self, v: Any) -> bool:
    return isinstance(v, int) and not isinstance(v, bool)


class QxFloatTyping(_QxCoreTyping):
  def check_typing(self, v: Any) -> bool:
    return isinstance(v, float)


class QxStrTyping(_QxCoreTyping):
  def check_typing(self, v: Any) -> bool:
    return isinstance(v, str)


class QxListTyping(QxTyping):
  def __init__(self, e: Optional[QxTyping] = None):
    self.e = e

  def check_typing(self, v: Any) -> bool:
    if isinstance(v, (tuple, list, set)):
      ret = True
      if isinstance(self.e, QxTyping):
        for i in v:
          if not self.e.check_typing(i):
            ret = False
            break
    else:
      ret = False
    return ret


class QxDictTyping(QxTyping):
  def __init__(self, k: Optional[QxTyping] = None, v: Optional[QxTyping] = None):
    self.k = k
    self.v = v

  def check_typing(self, v: Any) -> bool:
    if isinstance(v, dict):
      ret = True
      if isinstance(self.k, QxTyping) or isinstance(self.v, QxTyping):
        if isinstance(self.k, QxTyping):
          fk = lambda x: self.k.check_typing(x)
        else:
          fk = lambda x: True

        if isinstance(self.v, QxTyping):
          fv = lambda x: self.v.check_typing(x)
        else:
          fv = lambda x: True

        for ik, iv in v.items():
          if not (fk(ik) and fv(iv)):
            ret = False
            break
    else:
      ret = False
    return ret


class QxFunctionTyping(QxTyping):
  def __init__(self):
    super(QxFunctionTyping, self).__init__()

  def check_typing(self, v: Any) -> bool:
    return isinstance(v, QX_FUNCTION_TYPE)


class QxOptionalTyping(QxTyping):
  def __init__(self, e: QxTyping):
    self.e = e

  def check_typing(self, v: Any) -> bool:
    ret = False

    if v is None:
      ret = True
    elif isinstance(self.e, type) and isinstance(v, self.e):
      ret = True

    return ret


class QxUnionTyping(QxTyping):
  def __init__(self, el: List[QxTyping]):
    self.el = el

  def match(self, v: Any) -> QxTyping:
    ret = None

    for e in self.el:
      if e.check_typing(v):
        ret = e
        break

    return ret

  def check_typing(self, v: Any) -> bool:
    return isinstance(self.match(v), QxTyping)


class QxTypingCache(QxCache):
  _lock = threading.Lock()
  _inst = None

  @classmethod
  def __new__(cls, *a, **kw):
    if not isinstance(cls._inst, cls):
      with cls._lock:
        if not isinstance(cls._inst, cls):
          cls._inst = super().__new__(cls)
    return cls._inst

  def __init__(self):
    super(QxTypingCache, self).__init__(size=128)


class QxTypingUtils(object):
  @staticmethod
  def is_alias_typing(t: Any) -> bool:
    return str(t).startswith('typing.') or t.__class__.__name__ == 'TypeVar'

  @staticmethod
  def is_basic_typing(t: Any) -> bool:
    return t is None or isinstance(t, type)

  @staticmethod
  def parse_typing(t: Any) -> QxTyping:
    cache = QxTypingCache()

    ret = QxAnyTyping()

    if isinstance(t, QxTyping):
      ret = t
    elif QxTypingUtils.is_alias_typing(t):
      fullname = str(t)
      if fullname.startswith('typing.'):
        name = fullname[7:].split('[', maxsplit=1)[0]
      else:
        name = t.__class__.__name__

      origin = getattr(t, '__origin__', None)
      args = getattr(t, '__args__', ())

      args_i = lambda i: QxTypingUtils.parse_typing(args[i]) if isinstance(args, (tuple, list)) and len(args) >= i + 1 else None
      args_t = lambda: [QxTypingUtils.parse_typing(arg) for arg in args] if isinstance(args, (tuple, list)) else []

      ret = cache.get(fullname)
      if not isinstance(ret, QxTyping):
        if name == 'Optional':
          ret = QxOptionalTyping(e=args_i(0))
        elif name == 'Union':
          ret = QxUnionTyping(el=args_t())
        elif name in ('List', 'Tuple'):
          ret = QxListTyping(e=args_i(0))
        elif name == 'Dict':
          ret = QxDictTyping(k=args_i(0), v=args_i(1))
        elif name == 'Text':
          ret = QxStrTyping()
        elif name == 'TypeVar':
          ret = QxAnyTyping()
        elif name in ('Callable', 'Function'):
          ret = QxFunctionTyping()
        else:
          ret = QxTypingUtils.parse_typing(origin)
        cache.set(fullname, ret)

    elif QxTypingUtils.is_basic_typing(t):
      if t is None or isinstance(t, type) and isinstance(None, t):
        ret = QxNoneTyping()
      elif t is bool:
        ret = QxBoolTyping()
      elif t is int:
        ret = QxIntTyping()
      elif t is float:
        ret = QxFloatTyping()
      elif t is str:
        ret = QxStrTyping()
      elif t is list or t is tuple:
        ret = QxListTyping()
      elif t is dict:
        ret = QxDictTyping()

    return ret

  @staticmethod
  def check_typing(v: Any, t: Any) -> bool:
    return QxTypingUtils.parse_typing(t).check_typing(v)

  @staticmethod
  def assert_typing(v: Any, t: Any) -> NoReturn:
    assert QxTypingUtils.parse_typing(t).assert_typing(v)



from typing import *

from urllib.parse import urlparse, quote, unquote

import os
from collections import OrderedDict


class QxUrl(object):
  def __init__(self, url: Union[str, object] = None, **kw):
    self._protocol = None
    self._username = None
    self._password = None
    self._hostname = None
    self._port = None
    self._path = None
    self._query = None
    self._query_str = None
    self._fragment = None

    self.set_url(url)

  @property
  def protocol(self) -> str:
    return self._protocol

  @protocol.setter
  def protocol(self, protocol: str):
    self._protocol = protocol

  @property
  def username(self) -> str:
    return self._username

  @username.setter
  def username(self, username: str):
    self._username = username

  @property
  def password(self) -> str:
    return self._password

  @password.setter
  def password(self, password: str):
    self._password = password

  @property
  def hostname(self) -> str:
    return self._hostname

  @hostname.setter
  def hostname(self, hostname: str):
    self._hostname = hostname

  @property
  def port(self) -> int:
    return self._port

  @port.setter
  def port(self, port: Union[int, str, None]):
    if isinstance(port, str):
      try:
        port = int(port)
      except ValueError:
        port = None

    self._port = port if isinstance(port, int) and port >= 0 else None

  @property
  def path(self) -> str:
    return self._path

  @path.setter
  def path(self, path: str):
    if not path.startswith('/'):
      path = f'/{path}'
    self._path = path

  @staticmethod
  def parse_query_str(query_str: str) -> dict:
    q = OrderedDict()
    for kv in query_str.split('&'):
      if not len(kv.strip()):
        continue

      kv = kv.split('=', maxsplit=1)

      k = kv[0]
      v = unquote(kv[1]) if len(kv) > 1 else ''

      if k in q:
        if isinstance(q[k], (tuple, list)):
          q[k] = list(q[k]) + [v]
        else:
          q[k] = [q[k], v]
      else:
        q[k] = v
    return q

  @staticmethod
  def mk_query_str(query: dict) -> str:
    items = []
    for k, vl in query.items():
      if not isinstance(vl, (tuple, list)):
        vl = [vl]
      for v in vl:
        items.append(f'{k}={quote(v)}')
    return '&'.join(items)

  @property
  def query(self) -> dict:
    return self._query

  @property
  def query_str(self) -> str:
    return self._query_str

  @query.setter
  def query(self, query: Union[str, dict]):
    if isinstance(query, str):
      query = self.parse_query_str(query)
      query_str = self.mk_query_str(query)
    elif isinstance(query, dict):
      query_str = self.mk_query_str(query)
      query = self.parse_query_str(query_str)
    else:
      query = {}
      query_str = ''
    self._query = query
    self._query_str = query_str

  @property
  def path_with_query(self) -> str:
    if self.query_str:
      ret = f'{self.path}?{self.query_str}'
    else:
      ret = self.path
    return ret

  @property
  def fragment(self) -> str:
    return self._fragment

  @fragment.setter
  def fragment(self, fragment: str):
    self._fragment = fragment

  def set_protocol(self, protocol: str):
    self.protocol = protocol
    return self

  def set_username(self, username: str):
    self.username = username
    return self

  def set_password(self, password: str):
    self.password = password
    return self

  def set_hostname(self, hostname: str):
    self.hostname = hostname
    return self

  def set_port(self, port: Union[int, str, None]):
    self.port = port
    return self

  def set_path(self, path: str):
    self.path = path
    return self

  def join_path(self, *paths):
    self.path = os.path.join('/' if not self.path else self.path, *paths)
    return self

  def set_query(self, query: Union[str, dict]):
    self.query = query
    return self

  def set_param(self, k, v, arr=False):
    v = unquote(v)
    q = self._query
    if arr and k in self._query:
      if isinstance(q[k], (tuple, list)):
        q[k] = list(q[k]) + [v]
      else:
        q[k] = [q[k], v]
    else:
      q[k] = v
    self._query = q
    self._query_str = self.mk_query_str(q)
    return self

  def set_fragment(self, fragment: str):
    self.fragment = fragment
    return self

  def set_url(self, url: str):
    if isinstance(url, str):
      p = urlparse(url)
      self.protocol = p.scheme
      self.username = p.username
      self.password = p.password
      self.hostname = p.hostname
      self.port = p.port
      self.path = p.path
      self.query = p.query
      self.fragment = p.fragment
    elif isinstance(url, QxUrl):
      self._protocol = url._protocol
      self._username = url._username
      self._password = url._password
      self._hostname = url._hostname
      self._port = url._port
      self._path = url._path
      self._query = url._query
      self._query_str = url._query_str
      self._fragment = url._fragment
    return self

  def mk_url(self) -> str:
    s = f'{self.protocol}://'
    if self.username:
      s += quote(self.username)
      if self.password:
        s += f':{quote(self.password)}'
      s += '@'
    s += self.hostname
    if self.port:
      s += f':{self.port}'
    s += quote(self.path)

    if self.query_str:
      s += f'?{self.query_str}'

    if self.fragment:
      s += f'#{quote(self.fragment)}'

    return s

  def __str__(self):
    return self.mk_url()

  def mk_dict(self) -> dict:
    return {
      'protocol': self.protocol,
      'username': self.username,
      'password': self.password,
      'hostname': self.hostname,
      'port': self.port,
      'query': self.query,
      'query_str': self.query_str,
      'fragment': self.fragment,
    }



import re

PATTERN_NONE = re.compile(r'NONE|NULL|NIL|', re.I)
PATTERN_TRUE = re.compile(r'1|TRUE|T|YES|Y|ON', re.I)
PATTERN_FALSE = re.compile(r'0|FALSE|F|NO|N|OFF', re.I)
PATTERN_BOOL = re.compile(rf'({PATTERN_TRUE.pattern})|({PATTERN_FALSE.pattern})', re.I)
PATTERN_BIN = re.compile(r'[-+]?0?b[01]+', re.I)
PATTERN_OCT = re.compile(r'[-+]?0[0-7]+', re.I)
PATTERN_HEX = re.compile(r'[-+]?0x[\dA-F]+', re.I)
PATTERN_INT = re.compile(r'[-+]?(0|[1-9]\d*)(\.0*)?(e\d+)?', re.I)
PATTERN_UINT = re.compile(r'[+]?(0|[1-9]\d*)(\.0*)?(e\d+)?', re.I)
PATTERN_FLOAT = re.compile(r'[-+]?(0|[1-9]\d*)(\.\d*)?(e\d+)?', re.I)
PATTERN_BASE_ENC = re.compile(r'base(16|32|64|85),.*', re.I)
PATTERN_MD5 = re.compile(r'[\dA-F]{32}', re.I)
PATTERN_SHA1 = re.compile(r'[\dA-F]{40}', re.I)
PATTERN_SHA256 = re.compile(r'[\dA-F]{64}', re.I)
PATTERN_SHA512 = re.compile(r'[\dA-F]{128}', re.I)
PATTERN_UUID = re.compile(r'([\dA-F]{8})-?([\dA-F]{4})-?([\dA-F]{4})-?([\dA-F]{4})-?([\dA-F]{12})', re.I)
PATTERN_DATE = re.compile(r'([0-2]\d{3})[-._年]?(0?[1-9]|1[0-2])[-._月]?(0?[1-9]|[1-2]\d|3[01])[日]?', re.I)
PATTERN_TIME = re.compile(r'([01]\d|2[0-3])[-._:时]?([0-5]\d)[-._:分]?([0-5]\d)[秒]?', re.I)
PATTERN_DATETIME = re.compile(rf'({PATTERN_DATE.pattern})\s*({PATTERN_TIME.pattern})', re.I)
PATTERN_MAC = re.compile(r'[\dA-F]{2}([-:]?[\dA-F]{2}){5}', re.I)
PATTERN_IPV4 = re.compile(r'([01]\d\d|2[0-4]\d|25[0-5])(\.[01]\d\d|2[0-4]\d|25[0-5]){3}', re.I)
PATTERN_DOMAIN = re.compile(r'([\w-]+\.)+[\w-]+', re.I)
PATTERN_URL_USER = re.compile(r'[^:@.]+(:[^@.]+)?', re.I)
PATTERN_URL_HOST = re.compile(rf'(({PATTERN_IPV4.pattern})|({PATTERN_DOMAIN.pattern}))(:\d+)?', re.I)
PATTERN_URL_PATH = re.compile(r'/([^/?#]+/?)*', re.I)
PATTERN_URL_PATH_VARIABLE = re.compile(r'\{[^{}]*\}', re.I)
PATTERN_URL_QUERY = re.compile(r'[^=&#]+(=[^&#]+)?(&[^=&#]+(=[^&#]+)?)*', re.I)
PATTERN_URL_FRAGMENT = re.compile(r'#.*', re.I)
PATTERN_URL_NO_PROTO = re.compile(
  rf'({PATTERN_URL_USER.pattern}@)?{PATTERN_URL_HOST.pattern}({PATTERN_URL_PATH.pattern})?({PATTERN_URL_QUERY.pattern})?({PATTERN_URL_FRAGMENT.pattern})?',
  re.I
)
PATTERN_URL_HTTP = re.compile(rf'HTTPS?://{PATTERN_URL_NO_PROTO.pattern}', re.I)
PATTERN_URL_GIT = re.compile(rf'GIT@{PATTERN_URL_HOST.pattern}:([^/]+){PATTERN_URL_PATH.pattern}\.GIT', re.I)
PATTERN_EMAIL = re.compile(rf'[^@]+@{PATTERN_URL_HOST.pattern}', re.I)
PATTERN_PHONE = re.compile(r'([08]\d{2,4})?\d{3,8}', re.I)
PATTERN_MOBILE = re.compile(r'([+]?0{0,2}86[-.\s]?)?(1\d{10})', re.I)
PATTERN_POSTCODE = re.compile(r'\d{6}', re.I)
PATTERN_EMOJI = re.compile(r'(\ud83c[\udf00-\udfff])|(\ud83d[\udc00-\ude4f\ude80-\udeff])|[\u2600-\u2B55]', re.I)
PATTERN_EN = re.compile(r'[A-Z]', re.I)
PATTERN_CN = re.compile(r'[\u4E00-\u9FFF]', re.I)
PATTERN_JP = re.compile(r'[\u0800-\u4DFF]', re.I)
PATTERN_KR = re.compile(r'[\uAC00-\uD7FF]', re.I)



from typing import *

from collections import OrderedDict


class QxCache(object):
  def __init__(self, size: int = 1024):
    self._size = size
    self._pool = OrderedDict()

  @property
  def size(self) -> int:
    return self._size

  @property
  def pool(self) -> OrderedDict:
    return self._pool

  def _refresh(self, k: str):
    if k in self.pool:
      self.pool.move_to_end(k, last=False)

  def _gc(self):
    if isinstance(self.size, int) and self.size > 0:
      while len(self.pool) > self.size:
        self.pool.popitem(last=True)

  def set(self, k: str, v: Any) -> None:
    self.pool[k] = v
    self._refresh(k)
    self._gc()

  def get(self, k: str, default: Any = None) -> Any:
    self._refresh(k)
    return self.pool.get(k, default)

  def has(self, k: str) -> bool:
    return k in self.pool

  def clear(self) -> None:
    self.pool.clear()

  def __contains__(self, k: str):
    return self.has(k)

  def __setitem__(self, k: str, v: Any):
    self.set(k, v)

  def __getitem__(self, k: str) -> Any:
    return self.get(k)

  def __str__(self):
    return f'{self.__class__.__name__}{str(self.pool)[len(self.pool.__class__.__name__):]}'



from typing import *

import os
import json
import threading


class QxConf(dict):
  _ns = None
  _parent = None

  def __init__(self, ns: str = None, parent: object = None, *a, **kw):
    super(QxConf, self).__init__(*a, **kw)

    self.ns = ns
    self.parent = parent

  @property
  def ns(self) -> str:
    return self._ns

  @ns.setter
  def ns(self, ns: str):
    self._ns = ns.upper() if isinstance(ns, str) else ''

  @property
  def parent(self):
    return self._parent  # type: QxConf

  @parent.setter
  def parent(self, parent: object = None):
    self._parent = parent if isinstance(parent, QxConf) else None

  def update_from_json(self, json_str: str):
    try:
      self.update(json.loads(json_str))
    except (json.JSONDecodeError, TypeError, ValueError):
      pass
    return self

  def update_from_file(self, path: str, encoding: str = 'utf-8'):
    if os.path.isfile(path):
      with open(path, 'r+', encoding=encoding) as fo:
        self.update_from_json(fo.read())
    return self

  def write_to_file(self, path: str, encoding: str = 'utf-8'):
    d = os.path.dirname(path)
    if not os.path.isdir(d):
      os.makedirs(d)
    with open(path, 'w+', encoding=encoding) as fo:
      fo.write(json.dumps(self, indent=2, ensure_ascii=False))

  def pprint(self):
    print(json.dumps(self, indent=2, ensure_ascii=False))

  @staticmethod
  def mk_env_k(*items):
    return '_'.join([x.strip() for x in items if isinstance(x, str) and x.strip()]).upper()

  def get(self, k, default: Any = None, check: Callable = None) -> Any:
    if isinstance(self.parent, QxConf):
      default = self.parent.get(k, default=default, check=check)

    v = super(QxConf, self).get(k, os.environ.get(self.mk_env_k(self.ns, k), default))

    if callable(check) and not check(v):
      v = default

    if isinstance(v, dict):
      v = QxConf(self.mk_env_k(self.ns, k), v.items())

    return v

  @staticmethod
  def is_true(v: Any) -> bool:
    if isinstance(v, bool):
      ret = v is True
    elif isinstance(v, str):
      ret = v.upper() in ('TRUE', 'T', 'YES', 'Y', 'ON')
    elif isinstance(v, int):
      ret = v == 1
    else:
      ret = False
    return ret

  @staticmethod
  def is_false(v: Any) -> bool:
    if isinstance(v, bool):
      ret = v is False
    elif isinstance(v, str):
      ret = v.upper() in ('FALSE', 'F', 'NO', 'N', 'OFF')
    elif isinstance(v, int):
      ret = v == 0
    else:
      ret = False
    return ret

  @staticmethod
  def is_not_true(v: Any) -> bool:
    return not QxConf.is_true(v)

  @staticmethod
  def is_not_false(v: Any) -> bool:
    return not QxConf.is_false(v)


class QxGlobalConf(QxConf):
  _lock = threading.Lock()
  _inst = None

  def __new__(cls, *a, **kw):
    if not isinstance(cls._inst, cls):
      with cls._lock:
        if not isinstance(cls._inst, cls):
          cls._inst = super().__new__(cls)
    return cls._inst

  def __init__(self):
    super(QxGlobalConf, self).__init__('QX')
    self._root = os.path.join(os.path.dirname(os.path.abspath(__file__)))

    self._i18n = QxConf()
    self._i18n.update_from_file(os.path.join(self._root, 'conf', 'i18n', 'zh-cn.json'))
    self._i18n.update_from_file(os.path.join(os.getcwd(), 'conf', 'i18n', 'zh-cn.json'))

    self._code = QxConf()

    code = QxConf()
    code.update_from_file(os.path.join(self._root, 'conf', 'code.json'))
    code.update_from_file(os.path.join(os.getcwd(), 'conf', 'code.json'))

    for k, v in code.items():
      k = int(k)
      v = str(v)
      self._code[k] = v
      self._code[v] = k

    self.update_from_file(os.path.join(self._root, 'conf', 'conf.json'))
    self.update_from_file(os.path.join(os.getcwd(), 'conf', 'conf.json'))

  @property
  def i18n(self):
    return self._i18n

  @property
  def code(self):
    return self._code

  @property
  def charset(self):
    return self.get('charset', 'utf-8')

  @property
  def is_debug(self) -> bool:
    return self.is_true(self.get('debug'))



from typing import *

import re
import uuid
import time
import redis
import random
from .value import *
from datetime import date, datetime, timedelta
from .decorator import *


class QxRedisClient(object):
  __connection_pool_cache__ = dict()

  def __init__(self, host: str = '127.0.0.1', port: int = 6379, password: Optional[str] = None, database: int = 0, current_time: Optional[datetime] = None):
    if isinstance(host, str):
      if '://' in host:
        conn_pool = QxRedisClient.create_conn_pool_from_url(host)
      else:
        conn_pool = QxRedisClient.create_conn_pool(host, port, password, database)
    else:
      assert isinstance(host, redis.ConnectionPool)
      conn_pool = host
    self._redis_db = redis.Redis(connection_pool=conn_pool)
    self._current_time = None

    self.current_time = current_time

  @classmethod
  def create_conn_pool_from_url(cls, url: str) -> redis.ConnectionPool:
    cache_id = url
    if cache_id in cls.__connection_pool_cache__:
      conn_pool = cls.__connection_pool_cache__[cache_id]
    else:
      conn_pool = redis.BlockingConnectionPool.from_url(url)
      cls.__connection_pool_cache__.setdefault(cache_id, conn_pool)
    return conn_pool

  @classmethod
  def create_conn_pool(cls, host: str = 'localhost', port: int = 6379, password: Optional[str] = None, database: int = 0) -> redis.ConnectionPool:
    if password:
      url = 'redis://{}@{}:{}/{}'.format(password, host, port, database)
    else:
      url = 'redis://{}:{}/{}'.format(host, port, database)
    return cls.create_conn_pool_from_url(url)

  @property
  def redis_db(self) -> redis.Redis:
    return self._redis_db

  @property
  def current_time(self) -> datetime:
    return self._current_time

  @current_time.setter
  def current_time(self, v: Optional[datetime]):
    self._current_time = v if isinstance(v, datetime) else datetime.now()

  @staticmethod
  def encode_k(k: str) -> str:
    return str(k)

  @staticmethod
  def decode_k(k: str) -> str:
    return k

  @staticmethod
  def encode_v(v: Any) -> bytes:
    return pickle.dumps(v)

  @staticmethod
  def decode_v(v: Optional[bytes]) -> Any:
    return v if v is None else pickle.loads(v)

  def ttl(self, name: str):
    name = self.encode_k(name)
    return self.redis_db.ttl(name)

  def pttl(self, name: str):
    name = self.encode_k(name)
    return self.redis_db.pttl(name)

  def expire(self, name: str, ttl: Union[Callable, date, datetime, int, float], at_least: bool = True):
    name = self.encode_k(name)

    ttl_old = self.redis_db.ttl(name)

    if callable(ttl):
      ttl = ttl()

    if isinstance(ttl, date):
      ttl_new = ttl if isinstance(ttl, datetime) else datetime(year=ttl.year, month=ttl.month, day=ttl.day)
      self.redis_db.expireat(name, ttl_new)
    else:
      if isinstance(ttl, timedelta):
        ttl_new = int(ttl.total_seconds())
      elif isinstance(ttl, (int, float)):
        ttl_new = int(ttl)
      else:
        ttl_new = ttl_old

      ttl_new = max(ttl_new, ttl_old) if at_least else ttl_new

      self.redis_db.expire(name, ttl_new)

  def exists(self, name: str):
    name = self.encode_k(name)
    return self.redis_db.exists(name)

  def setnx(self, name: str, value: Any):
    name = self.encode_k(name)
    value = self.encode_v(value)
    return self.redis_db.setnx(name, value)

  def set(self, name: str, value: Any, ex: Optional[int] = None, px: Optional[int] = None, nx: bool = False, xx: bool = False):
    name = self.encode_k(name)
    value = self.encode_v(value)
    return self.redis_db.set(name, value, ex, px, nx, xx)

  def get(self, name: str, default: Any = None):
    name = self.encode_k(name)
    value = default
    if self.redis_db.exists(name):
      value = self.redis_db.get(name)
      value = self.decode_v(value)
    return value

  def delete(self, name: str):
    name = self.encode_k(name)
    return self.redis_db.delete(name)

  @rhe
  def hexists(self, name: str, key: str):
    name = self.encode_k(name)
    key = self.encode_k(key)
    return self.redis_db.hexists(name, key)

  def hsetnx(self, name: str, key: str, value: Any):
    name = self.encode_k(name)
    key = self.encode_k(key)
    value = self.encode_v(value)
    return self.redis_db.hsetnx(name, key, value)

  @rhs
  def hset(self, name: str, key: str, value: Any):
    name = self.encode_k(name)
    key = self.encode_k(key)
    value = self.encode_v(value)
    return self.redis_db.hset(name, key, value)

  @rhg
  def hget(self, name: str, key: str):
    name = self.encode_k(name)
    key = self.encode_k(key)
    value = self.redis_db.hget(name, key)
    value = self.decode_v(value)
    return value

  def sadd(self, name: str, value: Any):
    name = self.encode_k(name)
    value = self.encode_v(value)
    return self.redis_db.sadd(name, value)

  def srem(self, name: str, value: Any):
    name = self.encode_k(name)
    value = self.encode_v(value)
    return self.redis_db.srem(name, value)

  def smembers(self, name: str):
    name = self.encode_k(name)
    values = self.redis_db.smembers(name)
    values = list(values)
    for i in range(len(values)):
      values[i] = self.decode_v(values[i])
    return values

  def scard(self, name: str):
    name = self.encode_k(name)
    count = self.redis_db.scard(name)
    return count

  def hdel(self, name: str, key: str):
    name = self.encode_k(name)
    key = self.encode_k(key)
    return self.redis_db.hdel(name, key)

  def keys(self, pattern: str, re_pattern: Optional[Union[str, re.Pattern]] = None):
    assert isinstance(pattern, str)
    keys = self.redis_db.keys(pattern)
    if re_pattern is not None:
      assert isinstance(re_pattern, (re.Pattern, str))
      keys, keys_t = [], keys
      if isinstance(re_pattern, str):
        re_pattern = re.compile(re_pattern)
      for k in keys_t:
        if re_pattern.match(k):
          keys.append(k)
    return keys

  def hgetall(self, name: str):
    d = dict()
    hd = self.redis_db.hgetall(name)
    if isinstance(hd, dict):
      for (k, v) in hd.items():
        k = self.decode_k(k)
        v = self.decode_v(v)
        d[k] = v
    return d

  def register_script(self, script: str):
    return self.redis_db.register_script(script)

  def pipeline(self, transaction: bool = True, shard_hint=None):
    return self.redis_db.pipeline(transaction, shard_hint)

  def blpop(self, keys: List[str], timeout: int = 0):
    return self.redis_db.blpop(keys, timeout)

  def brpop(self, keys: List[str], timeout: int = 0):
    return self.redis_db.brpop(keys, timeout)

  def brpoplpush(self, src: str, dst: str, timeout: int = 0):
    return self.redis_db.brpoplpush(src, dst, timeout)

  def lindex(self, name: str, index: int):
    return self.redis_db.lindex(name, index)

  def linsert(self, name: str, where: str, refvalue: Any, value: Any):
    return self.redis_db.linsert(name, where, refvalue, value)

  def llen(self, name: str):
    return self.redis_db.llen(name)

  def lpop(self, name: str):
    return self.redis_db.lpop(name)

  def lpush(self, name: str, *values: List[Any]):
    return self.redis_db.lpush(name, *values)

  def lpushx(self, name: str, value: Any):
    return self.redis_db.lpushx(name, value)

  def lrange(self, name: str, start: int, end: int):
    return self.redis_db.lrange(name, start, end)

  def lrem(self, name: str, count: int, value: Any):
    return self.redis_db.lrem(name, count, value)

  def lset(self, name: str, index: int, value: Any):
    return self.redis_db.lset(name, index, value)

  def ltrim(self, name: str, start: int, end: int):
    return self.redis_db.ltrim(name, start, end)

  def rpop(self, name: str):
    return self.redis_db.rpop(name)

  def rpoplpush(self, src: str, dst: str):
    return self.redis_db.rpoplpush(src, dst)

  def rpush(self, name: str, *values: List[Any]):
    return self.redis_db.rpush(name, *values)

  def rpushx(self, name: str, value: Any):
    return self.redis_db.rpushx(name, value)

  def sort(self, name: str, start: int = None, num: int = None, by=None, get=None, desc=False, alpha=False, store=None, groups=False):
    return self.redis_db.sort(name, start, num, by, get, desc, alpha, store, groups)


class QxRedisLockError(Exception):
  pass


class QxRedisLock(object):
  DEFAULT_RETRY_TIMES = 3
  DEFAULT_RETRY_DELAY = 200
  DEFAULT_TTL = 100000
  CLOCK_DRIFT_FACTOR = 0.01
  RELEASE_LUA_SCRIPT = """
        if redis.call("get", KEYS[1]) == ARGV[1] then
            return redis.call("del", KEYS[1])
        else
            return 0
        end
    """

  def __init__(self, resource, redis_client, retry_times=DEFAULT_RETRY_TIMES, retry_delay=DEFAULT_RETRY_DELAY,
               ttl=DEFAULT_TTL):
    assert isinstance(redis_client, QxRedisClient)
    self._resource = resource
    self._redis_client = redis_client
    self._retry_times = retry_times
    self._retry_delay = retry_delay
    self._ttl = ttl
    self._lock_key = None
    self._redis_client = None

  @property
  def resource(self):
    return self._resource

  @property
  def redis_client(self):
    return self._redis_client

  @property
  def retry_times(self):
    return self._retry_times

  @property
  def retry_delay(self):
    return self._retry_delay

  @property
  def ttl(self):
    return self._ttl

  @property
  def lock_key(self):
    return self._lock_key

  @lock_key.setter
  def lock_key(self, v):
    self._lock_key = v

  def __enter__(self):
    if not self.acquire():
      raise QxRedisLockError(u'Failed to acquire RedisLock')

  def __exit__(self, exc_type, exc_value, traceback):
    self.release()

  @staticmethod
  def _total_ms(delta):
    delta_seconds = delta.seconds + delta.days * 24 * 3600
    return (delta.microseconds + delta_seconds * 10 ** 6) / 10 ** 3

  def acquire(self):
    is_acquired = False

    self._lock_key = 'Locks:{}'.format(uuid.uuid4())

    for retry in range(self.retry_times):
      start_time = datetime.utcnow()

      if self._redis_client.set(self.resource, self.lock_key, nx=True, px=self.ttl):
        is_acquired = True

      end_time = datetime.utcnow()
      elapsed_milliseconds = self._total_ms(end_time - start_time)

      drift = (self.ttl * QxRedisLock.CLOCK_DRIFT_FACTOR) + 2

      if is_acquired and self.ttl > (elapsed_milliseconds + drift):
        break
      else:
        self.release()
        time.sleep(random.randint(0, self.retry_delay) / 1000)
    return is_acquired

  def release(self):
    if not getattr(self._redis_client, '_release_script', None):
      self._redis_client._release_script = self._redis_client.register_script(QxRedisLock.RELEASE_LUA_SCRIPT)
    self._redis_client._release_script(keys=[self.resource], args=[self.lock_key])


class QxRedisReentrantLock(QxRedisLock):
  def __init__(self, *args, **kwargs):
    super(QxRedisReentrantLock, self).__init__(*args, **kwargs)
    self._acquired_count = 0

  @property
  def acquired_count(self):
    return self._acquired_count

  def acquire(self):
    is_acquired = True
    if self._acquired_count == 0:
      is_acquired = super(QxRedisReentrantLock, self).acquire()

    if is_acquired:
      self._acquired_count += 1

    return is_acquired

  def release(self):
    is_released = False

    if self._acquired_count > 0:
      self._acquired_count -= 1
      if self._acquired_count == 0:
        super(QxRedisReentrantLock, self).release()
        is_released = True

    return is_released

  def release_all(self):
    if self._acquired_count > 0:
      super(QxRedisReentrantLock, self).release()
    self._acquired_count = 0


class QxRedisVariable(object):
  def __init__(self, name, redis_client):
    assert isinstance(name, str)
    assert isinstance(redis_client, QxRedisClient)
    self._name = name
    self._redis_name = 'vars:{}'.format(name)
    self._redis_client = redis_client
    self._lock = QxRedisReentrantLock(resource=name, redis_client=redis_client)

  @property
  def name(self):
    return self._name

  @property
  def redis_name(self):
    return self._redis_name

  @property
  def redis_client(self):
    return self._redis_client

  @property
  def lock(self):
    return self._lock

  def get_value(self, default=None):
    with self.lock:
      v = self.redis_client.get(self.redis_name, default=default)
    return v

  def set_value(self, v, ttl=None, at_least=False):
    with self.lock:
      self.redis_client.set(self.redis_name, v)
      self.redis_client.expire(self.redis_name, ttl=ttl, at_least=at_least)

  def expire(self, ttl, at_least=False):
    with self.lock:
      self.redis_client.expire(self.redis_name, ttl=ttl, at_least=at_least)


class QxRedisInt(QxRedisVariable):
  def __int__(self):
    return int(self.get_value(0))

  def __add__(self, other):
    with self.lock:
      v = int(self.get_value(0))
      v += other
    return v

  def __iadd__(self, other):
    return self.inc_and_return(other)

  def inc_and_return(self, increment=1):
    with self.lock:
      v = int(self.get_value(0))
      v += increment
      self.set_value(v)
    return v

  def return_and_inc(self, increment=1):
    with self.lock:
      v = int(self.get_value(0))
      vv = v + increment
      self.set_value(vv)
    return v



import random
from typing import *

import copy
from uuid import uuid4 as uuu
from time import sleep as sss
from .inspectutils import QxInspectUtils as iu


class QxDecoratorHook(object):
  def __init__(self):
    self._al = []
    self._kw = {}
    self._func_check = None
    self._func_target = None

  @property
  def al(self) -> List:
    return self._al

  @property
  def kw(self) -> Dict:
    return self._kw

  def set_func_check(self, func: Callable):
    self._func_check = func

  def set_func_target(self, func: Callable):
    self._func_target = func

  def _check(self, target: Any) -> bool:
    raise NotImplementedError()

  def _target(self, target: Any) -> Any:
    raise NotImplementedError()

  def check(self, target: Any) -> bool:
    if callable(self._func_check):
      return self._func_check(self, target)
    return self._check(target)

  def set_args(self, *al, **kw):
    self._al = copy.copy(al)
    self._kw = copy.copy(kw)

  def arg(self, k: Union[int, str, List[Union[int, str]]], default: Optional[Any] = None) -> Any:
    ret = default
    if isinstance(k, int):
      if len(self.al) > k:
        ret = self.al[k]
    elif isinstance(k, str):
      if k in self.kw:
        ret = self.kw[k]
    elif isinstance(k, list):
      for kk in k:
        if isinstance(kk, int):
          if len(self.al) > kk:
            ret = self.al[kk]
            break
        elif isinstance(kk, str):
          if kk in self.kw:
            ret = self.kw[kk]
            break
    return ret

  def target(self, target: Any) -> Any:
    if callable(self._func_target):
      return self._func_target(self, target)
    return self._target(target)


class QxDecoratorDefaultHook(QxDecoratorHook):
  def _check(self, target: Any) -> bool:
    return True

  def _target(self, target: Any) -> Any:
    return target


def QxDecorator(hook_cls: Optional[Type[QxDecoratorHook]] = None, func_check: Optional[Callable] = None, func_target: Optional[Callable] = None):
  if isinstance(hook_cls, QxDecoratorHook):
    hook_cls = hook_cls.__class__
  elif callable(hook_cls):
    hook_cls = hook_cls

  if not (isinstance(hook_cls, type) and issubclass(hook_cls, QxDecoratorHook)) or hook_cls == QxDecoratorHook:
    hook_cls = QxDecoratorDefaultHook

  def __init__(*al, **kw):
    hook = hook_cls()
    hook.set_func_check(func_check)
    hook.set_func_target(func_target)

    def __target__(target):
      return hook.target(target)

    if len(al) == 1 and hook.check(al[0]) and len(kw) == 0:
      return __target__(al[0])

    hook.set_args(*al, **kw)

    return __target__

  return __init__


